#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import errno

import shutil
import tarfile

import datetime

from . import jsonfiles, string
from .errors import *
from .folder import Folder
from .simulation import Simulation

class Manager():
	'''
	Manage a simulations folder: add, delete, extract or update simulations, based on their settings.

	Parameters
	----------
	folder : Folder|string
		The folder to manage. Either a `Folder` instance or the path to the folder (used to create a `Folder` instance).

	readonly : bool
		If `True`, open an instance in "read only" mode. Running indicator file is ignored, but only extraction is allowed.

	Raises
	------
	ManagerAlreadyRunningError
		A Manager instance is already running.
	'''

	def __init__(self, folder, *, readonly = False):
		self._folder = folder if type(folder) is Folder else Folder(folder)

		self._simulations_list_dict = None

		self._readonly = readonly

		self._delete_running_indicator = False

		if not(self._readonly):
			if os.path.isfile(self._folder.running_manager_indicator_filename):
				self._delete_running_indicator = False
				raise ManagerAlreadyRunningError()

			# Add a file into the configuration folder to indicate a Manager instance is currently running
			with open(self._folder.running_manager_indicator_filename, 'w') as f:
				f.write(str(datetime.datetime.now()))

			self._delete_running_indicator = True

	def __enter__(self):
		'''
		Context manager so we can use `with` instead of manually calling `close()`.
		'''

		return self

	def __exit__(self, type, value, traceback):
		'''
		Ensure `close()` is called when exiting the context manager.
		'''

		self.close()

	def close(self):
		'''
		Delete the "running indicator".
		Should always be called before destroying the instance to ensure the indicator is deleted.
		'''

		if self._delete_running_indicator:
			os.unlink(self._folder.running_manager_indicator_filename)

	@property
	def folder(self):
		'''
		Return the `Folder` instance.

		Returns
		-------
		folder : Folder
			The instance used by the manager.
		'''

		return self._folder

	@property
	def _simulations_list(self):
		'''
		Return the simulations list.

		Returns
		-------
		simulations_list : dict
			A name: settings dictionary.
		'''

		if self._simulations_list_dict is None:
			try:
				self._simulations_list_dict = jsonfiles.read(self._folder.simulations_list_filename)

			except FileNotFoundError:
				self._simulations_list_dict = {}

		return self._simulations_list_dict

	def saveSimulationsList(self):
		'''
		Save the list of simulations.

		Raises
		------
		OperationNotAllowed
			The list cannot be written in read only mode.
		'''

		if self._readonly:
			raise OperationNotAllowed()

		jsonfiles.write(self._simulations_list, self._folder.simulations_list_filename)

	def compress(self, folder, simulation_name):
		'''
		Create an archive to store a simulation.

		Parameters
		----------
		folder : str
			Folder to compress.

		simulation_name : str
			Name to use for the archive.
		'''

		with tarfile.open(os.path.join(self._folder.simulations_folder, f'{simulation_name}.tar.bz2'), 'w:bz2') as tar:
			tar.add(folder, arcname = simulation_name)

		shutil.rmtree(folder)

	def uncompress(self, simulation_name, folder):
		'''
		Extract a simulation from an archive.

		Parameters
		----------
		simulation_name : str
			Name of the archive to extract.

		folder : str
			Folder into which the files must go.
		'''

		with tarfile.open(os.path.join(self._folder.simulations_folder, f'{simulation_name}.tar.bz2'), 'r:bz2') as tar:
			tar.extractall(path = self._folder.tmp_folder)

		if os.path.isdir(folder):
			os.rmdir(folder)

		shutil.move(os.path.join(self._folder.tmp_folder, simulation_name), folder)

	def getSimulationsNumber(self):
		'''
		Returns the total number of simulations stored in the folder.

		Returns
		-------
		n : int
			The number of simulations.
		'''

		return len(self._simulations_list)

	def settingsOf(self, simulation_id):
		'''
		Return the whole settings set of a simulation.

		Parameters
		----------
		simulation_id : str
			Identifier of the simulation, either its name or its hashed settings.

		Raises
		------
		SimulationNotFoundError
			The simulation does not exist in the list.

		Returns
		-------
		settings : dict
			Settings of the simulation.
		'''

		try:
			if len(simulation_id) == 32:
				settings_str = [infos['settings'] for infos in self._simulations_list.values() if infos['name'] == simulation_id][0]

			else:
				settings_str = self._simulations_list[simulation_id]['settings']

		except (IndexError, KeyError):
			raise SimulationNotFoundError(simulation_id)

		else:
			return string.toObject(settings_str)

	def exists(self, simulation):
		'''
		Check whether a simulation does exist in the folder.

		Parameters
		----------
		simulation : Simulation|dict
			The simulation to check the existence of.

		Returns
		-------
		exists : bool
			`True` if the simulation exists, `False` otherwise.
		'''

		simulation = Simulation.ensureType(simulation, self._folder)
		settings_hashed = string.hash(string.fromObject(simulation.settings))

		return (settings_hashed in self._simulations_list)

	def add(self, simulation, *, save_list = True):
		'''
		Add a simulation to the list.

		Parameters
		----------
		simulation : Simulation|dict
			The simulation to add.

		save_list : boolean
			`True` to save the simulations list, `False` otherwise.

		Raises
		------
		OperationNotAllowed
			Addition is not allowed in read only mode.

		SimulationFolderNotFoundError
			The folder indicated in the simulation does not exist.

		SimulationIntegrityCheckFailedError
			At least one integrity check failed.
		'''

		if self._readonly:
			raise OperationNotAllowed()

		simulation = Simulation.ensureType(simulation, self._folder)

		if not(os.path.isdir(simulation['folder'])):
			raise SimulationFolderNotFoundError(simulation['folder'])

		settings_str = string.fromObject(simulation.settings)
		settings_hashed = string.hash(settings_str)
		simulation_name = string.uniqueID()

		if not(self._folder.checkIntegrity(simulation)):
			raise SimulationIntegrityCheckFailedError(simulation['folder'])

		self.compress(simulation['folder'], simulation_name)

		self._simulations_list[settings_hashed] = {
			'name': simulation_name,
			'added': datetime.datetime.now().isoformat(sep = ' ', timespec = 'seconds'),
			'settings': settings_str
		}

		if save_list:
			self.saveSimulationsList()

	def addFromFolder(self, folder, *, save_list = True, settings_file = 'settings.json'):
		'''
		Add a simulation from its folder.

		Parameters
		----------
		folder : str
			Path to the folder to add.

		save_list : boolean
			`True` to save the simulations list, `False` otherwise.

		settings_file : str
			Name of the file containing the settings.

		Raises
		------
		OperationNotAllowed
			Addition is not allowed in read only mode.

		SimulationFolderNotFoundError
			The indicated folder does not exist.

		SimulationIntegrityCheckFailedError
			At least one integrity check failed.
		'''

		if self._readonly:
			raise OperationNotAllowed()

		if not(os.path.isdir(folder)):
			raise SimulationFolderNotFoundError(folder)

		settings_filepath = os.path.join(folder, settings_file)
		settings = jsonfiles.read(settings_filepath)

		os.unlink(settings_filepath)

		try:
			self.add({
				'folder': folder,
				'settings': settings
			}, save_list = save_list)

		except SimulationIntegrityCheckFailedError:
			jsonfiles.write(settings, settings_filepath)
			raise

	def delete(self, simulation, *, save_list = True):
		'''
		Delete a simulation.

		Parameters
		----------
		simulation : Simulation|dict
			The simulation to delete.

		save_list : boolean
			`True` to save the simulations list, `False` otherwise.

		Raises
		------
		OperationNotAllowed
			Deletion is not allowed in read only mode.

		SimulationNotFoundError
			The simulation does not exist in the list.
		'''

		if self._readonly:
			raise OperationNotAllowed()

		simulation = Simulation.ensureType(simulation, self._folder)
		settings_hashed = string.hash(string.fromObject(simulation.settings))

		if not(settings_hashed in self._simulations_list):
			raise SimulationNotFoundError(settings_hashed)

		simulation_name = self._simulations_list[settings_hashed]['name']

		os.unlink(os.path.join(self._folder.simulations_folder, f'{simulation_name}.tar.bz2'))
		del self._simulations_list[settings_hashed]

		if save_list:
			self.saveSimulationsList()

	def deleteFromFolder(self, folder, *, save_list = True, settings_file = 'settings.json'):
		'''
		Delete a simulation from its folder.

		Parameters
		----------
		folder : str
			Path to the folder to delete.

		save_list : boolean
			`True` to save the simulations list, `False` otherwise.

		settings_file : str
			Name of the file containing the settings.

		Raises
		------
		OperationNotAllowed
			Deletion is not allowed in read only mode.

		SimulationFolderNotFoundError
			The indicated folder does not exist.
		'''

		if self._readonly:
			raise OperationNotAllowed()

		if not(os.path.isdir(folder)):
			raise SimulationFolderNotFoundError(folder)

		settings = jsonfiles.read(os.path.join(folder, settings_file))

		self.delete({
			'folder': folder,
			'settings': settings
		}, save_list = save_list)

	def extract(self, simulation, *, settings_file = None):
		'''
		Extract a simulation.

		Parameters
		----------
		simulation : Simulation|dict
			The simulation to extract.

		settings_file : str
			Name of the file to create to store the simulation's settings.

		Raises
		------
		SimulationNotFoundError
			The simulation does not exist in the list.

		SimulationFolderAlreadyExistError
			The destination of extraction already exists.
		'''

		simulation = Simulation.ensureType(simulation, self._folder)
		settings_hashed = string.hash(string.fromObject(simulation.settings))

		if not(settings_hashed in self._simulations_list):
			raise SimulationNotFoundError(settings_hashed)

		if os.path.exists(simulation['folder']):
			raise SimulationFolderAlreadyExistError(simulation['folder'])

		simulation_name = self._simulations_list[settings_hashed]['name']

		destination_path = os.path.dirname(os.path.normpath(simulation['folder']))
		if destination_path and not(os.path.isdir(destination_path)):
			os.makedirs(destination_path)

		self.uncompress(simulation_name, simulation['folder'])

		if settings_file:
			simulation.writeSettingsFile(settings_file)

	def batchAction(self, simulations, action, args = {}, *, save_list = True, errors_store = (), errors_pass = (Error), callback = None):
		'''
		Apply a callback function to each simulation of a given list.

		Parameters
		----------
		simulations : list
			List of simulations.

		action : function
			Function to call. The simulation will be passed as the first parameter.

		args : dict
			Additional named arguments to pass to the callback.

		save_list : boolean
			`True` to save the simulations list once the loop is over, `False` to not save it.

		errors_store : tuple
			List of exceptions that, when raised, lead to the storage of the involved simulation.

		errors_pass : tuple
			List of exceptions that, when raised, do nothing.

		callback : function
			Function to call at the end of each action.

		Returns
		-------
		errors : list
			List of simulations which raised an error.
		'''

		errors = []

		for simulation in simulations:
			try:
				action(simulation, **args)

			except errors_store:
				errors.append(simulation)

			except errors_pass:
				pass

			if not(callback is None):
				callback()

		if save_list:
			self.saveSimulationsList()

		return errors

	def batchAdd(self, simulations, *, callback = None):
		'''
		Add multiple simulations to the list.

		Parameters
		----------
		simulations : list
			List of simulations.

		callback : function
			Function to call at the end of each addition.

		Returns
		-------
		errors : list
			List of simulations that were not added because they raised an error.
		'''

		return self.batchAction(simulations, self.add, {'save_list': False}, save_list = True, errors_store = (SimulationFolderNotFoundError, SimulationIntegrityCheckFailedError), callback = callback)

	def batchAddFromFolder(self, folders, *, settings_file = 'settings.json', callback = None):
		'''
		Add multiple simulations from their folders.

		Parameters
		----------
		folders : list
			List of folders.

		callback : function
			Function to call at the end of each addition.

		Returns
		-------
		errors : list
			List of simulations that were not added because they raised an error.
		'''

		return self.batchAction(folders, self.addFromFolder, {'settings_file': settings_file, 'save_list': False}, save_list = True, errors_store = (SimulationFolderNotFoundError, SimulationIntegrityCheckFailedError), callback = callback)

	def batchDelete(self, simulations, *, callback = None):
		'''
		Delete multiple simulations.

		Parameters
		----------
		simulations : list
			List of simulations.

		callback : function
			Function to call at the end of each deletion.

		Returns
		-------
		errors : list
			List of simulations that were not deleted because they raised an error.
		'''

		return self.batchAction(simulations, self.delete, {'save_list': False}, save_list = True, errors_store = (SimulationNotFoundError), callback = callback)

	def batchDeleteFromFolder(self, folders, *, settings_file = 'settings.json', callback = None):
		'''
		Delete multiple simulations from their folders.

		Parameters
		----------
		folders : list
			List of folders.

		callback : function
			Function to call at the end of each deletion.

		Returns
		-------
		errors : list
			List of simulations that were not deleted because they raised an error.
		'''

		return self.batchAction(folders, self.deleteFromFolder, {'settings_file': settings_file, 'save_list': False}, save_list = True, errors_store = (SimulationNotFoundError), callback = callback)

	def batchExtract(self, simulations, *, settings_file = None, ignore_existing = True, callback = None):
		'''
		Extract multiple simulations.

		Parameters
		----------
		simulations : list
			List of simulations.

		ignore_existing : boolean
			Ignore simulations for which the destination folder already exists.

		callback : function
			Function to call at the end of each extraction.

		Returns
		-------
		errors : list
			List of simulations that were not extracted because they raised an error.
		'''

		if ignore_existing:
			errors_store = (SimulationNotFoundError,)
			errors_pass = (SimulationFolderAlreadyExistError,)

		else:
			errors_store = (SimulationNotFoundError, SimulationFolderAlreadyExistError)
			errors_pass = ()

		return self.batchAction(simulations, self.extract, {'settings_file': settings_file}, save_list = False, errors_store = errors_store, errors_pass = errors_pass, callback = callback)

	def update(self, callback = None):
		'''
		Update the simulations list to take into account new settings.
		If the archive no longer exists, remove it from the simulations list.

		Parameters
		----------
		callback : function
			Function to call at each treated simulation.

		Raises
		------
		OperationNotAllowed
			Updating is not allowed in read only mode.
		'''

		if self._readonly:
			raise OperationNotAllowed()

		new_simulations_list = {}

		for settings_hashed, infos in self._simulations_list.items():
			if os.path.isfile(os.path.join(self._folder.simulations_folder, f'{infos["name"]}.tar.bz2')):
				simulation = Simulation.ensureType({
					'folder': '',
					'settings': string.toObject(infos['settings'])
				}, self._folder)

				settings_str = string.fromObject(simulation.settings)
				settings_hashed = string.hash(settings_str)

				new_simulations_list[settings_hashed] = {
					**infos,
					**{'settings': settings_str}
				}

			if not(callback is None):
				callback()

		self._simulations_list_dict = new_simulations_list
		self.saveSimulationsList()

	def transform(self, transformation, simulations_settings = None, callback = None):
		'''
		Apply a transformation operation to a given list of simulations, e.g. to store new informations.

		Parameters
		----------
		transformation : function
			Transformation to apply. This function must accept the following parameters.

				simulation : Simulation
					A Simulation object representing the settings of the current simulation.

			The function returns the new settings of the simulation.
			If `None`, the settings are not changed.

		simulations_settings : list
			List of simulations to transform (only their settings). If `None`, will transform all the stored simulations.

		callback : function
			Function to call once a simulation has been transformed.

		Raises
		------
		OperationNotAllowed
			Transformation is not allowed in read only mode.
		'''

		if self._readonly:
			raise OperationNotAllowed()

		if not(simulations_settings):
			simulations_settings = [string.toObject(infos['settings']) for infos in self._simulations_list.values()]

		for settings in simulations_settings:
			simulation_dir = self._folder.tempdir()

			simulation = Simulation.ensureType({
				'folder': simulation_dir,
				'settings': settings
			}, self._folder)

			settings_hashed = string.hash(string.fromObject(simulation.settings))
			simulation_infos = self._simulations_list[settings_hashed]

			self.uncompress(simulation_infos['name'], simulation_dir)
			new_settings = transformation(simulation)
			os.unlink(os.path.join(self._folder.simulations_folder, f'{simulation_infos["name"]}.tar.bz2'))
			self.compress(simulation_dir, simulation_infos['name'])

			if not(new_settings is None):
				new_simulation = Simulation.ensureType({
					'folder': simulation_dir,
					'settings': new_settings
				}, self._folder)

				new_settings_str = string.fromObject(new_simulation.settings)
				new_settings_hashed = string.hash(new_settings_str)

				del self._simulations_list[settings_hashed]

				self._simulations_list[new_settings_hashed] = {
					**simulation_infos,
					**{'settings': new_settings_str}
				}

				self.saveSimulationsList()

			if not(callback is None):
				callback()

	def clear(self, callback = None):
		'''
		Delete all simulations in the folder.

		Parameters
		----------
		callback : function
			Function to call at each deleted simulation.

		Raises
		------
		OperationNotAllowed
			Clearing is not allowed in read only mode.
		'''

		if self._readonly:
			raise OperationNotAllowed()

		for infos in self._simulations_list.values():
			os.unlink(os.path.join(self._folder.simulations_folder, f'{infos["name"]}.tar.bz2'))

			if not(callback is None):
				callback()

		self._simulations_list_dict = {}
		self.saveSimulationsList()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import copy
import functools
import re
import abc

from .errors import *
from . import string, jsonfiles

class Simulation():
	'''
	Represent a simulation, itself identified by its settings.

	Parameters
	----------
	folder : Folder
		Instance of `Folder` for the current simulation's folder.

	settings : dict
		Dictionary listing the user settings.
	'''

	def __init__(self, folder, settings):
		self._folder = folder
		self._user_settings = settings

		self._raw_globalsettings = None
		self._raw_settings = None

		self._indexed_settings = None

		self._setting_tag_regex_compiled = None
		self._eval_tag_regex_compiled = None
		self._parser_recursion_stack = []

	@classmethod
	def ensureType(cls, simulation, folder):
		'''
		Ensure a variable is a Simulation object.

		Parameters
		----------
		simulation : dict|Simulation
			The variable to check.

		folder : Folder
			The `Folder` instance to use in case we need to create a new object.

		Returns
		-------
		simulation : Simulation
			The simulation as a Simulation instance.
		'''

		if type(simulation) is cls:
			return simulation

		return cls(folder, simulation)

	def copy(self):
		'''
		Create a copy of the simulation.

		Returns
		-------
		simulation : Simulation
			The copy of the simulation.
		'''

		return Simulation(self._folder, {
			**self.globalsettings,
			'settings': {
				settings_set_name: [
					{s.name: s._value for s in settings_set}
					for settings_set in settings_sets
				]
				for settings_set_name, settings_sets in self._settings.items()
			}
		})

	@property
	def folder(self):
		'''
		Get the Folder instance used by the simulation.

		Returns
		-------
		folder : Folder
			The Folder instance.
		'''

		return self._folder

	def __getitem__(self, key):
		'''
		Access to a global setting.

		Parameters
		----------
		key : str
			The key of the setting to get.

		Raises
		------
		KeyError
			The key does not exist.

		Returns
		-------
		value : mixed
			The corresponding value.
		'''

		try:
			return self.globalsettings[key]

		except KeyError:
			raise KeyError('The key does not exist in the global settings')

	def __setitem__(self, key, value):
		'''
		Change a global setting.

		Parameters
		----------
		key : str
			The key of the setting to change.

		value : mixed
			The new value of the setting.

		Raises
		------
		KeyError
			The key does not exist.
		'''

		try:
			setting = [setting for setting in self._globalsettings if setting.name == key][0]

		except IndexError:
			raise KeyError('The key does not exist in the global settings')

		else:
			setting.value = value

	def writeSettingsFile(self, filename, *, folder = None):
		'''
		Write the settings into a file in the simulation's folder.

		Parameters
		----------
		filename : str
			Name of the file to write.

		folder : str
			Name of the folder in which the file should be saved. Default to the simulation's folder.
		'''

		jsonfiles.write(self.settings, os.path.join(folder or self['folder'], filename))

	@property
	def _globalsettings(self):
		'''
		Return (and generate if needed) the complete list of global settings.

		Returns
		-------
		raw_globalsettings : list
			The global settings.
		'''

		if self._raw_globalsettings is None:
			self.generateGlobalSettings()

		return self._raw_globalsettings

	@property
	def _settings(self):
		'''
		Return (and generate if needed) the complete list of settings as a dictionary.

		Returns
		-------
		raw_settings : dict
			The settings.
		'''

		if self._raw_settings is None:
			self.generateSettings()

		return self._raw_settings

	@property
	def raw_settings(self):
		'''
		Return a dictionary with the complete list of sets of settings.

		Returns
		-------
		settings : dict
			The list of settings.
		'''

		return {
			settings_set_name: [
				{s.name: s for s in settings_set}
				for settings_set in settings_sets
			]
			for settings_set_name, settings_sets in self._settings.items()
		}

	def getSetting(self, coords):
		'''
		Retrieve a setting from its set and name.

		Parameters
		----------
		coords : dict
			Describe the "coordinates" of the setting. Must contain the following keys:
				* `set`: the name of the set the setting belongs to,
				* `name`: the name of the setting.
			The following key is optional:
				* `set_index`: the index of the set. If not provided, default to 0.

		Returns
		-------
		setting : SimulationSetting
			The setting corresponding to the coordinates.
		'''

		coords = {
			'set_index': 0,
			**coords
		}

		return self.raw_settings[coords['set']][coords['set_index']][coords['name']]

	@property
	def settings(self):
		'''
		Return a dictionary with the complete list of sets of settings to use, as dictionaries.
		The settings with `exclude` to `True` are ignored.

		Returns
		-------
		settings : dict
			List of sets of settings.
		'''

		return {
			settings_set_name: [
				{s.name: s.value for s in settings_set if not(s.exclude)}
				for settings_set in settings_sets
			]
			for settings_set_name, settings_sets in self._settings.items()
		}

	@property
	def raw_values_settings(self):
		'''
		Return a dictionary with the complete list of sets of settings to use, as dictionaries.
		The settings with `exclude` to `True` are ignored.

		Returns
		-------
		settings : dict
			List of sets of settings.
		'''

		return {
			settings_set_name: [
				{s.name: s._value for s in settings_set}
				for settings_set in settings_sets
			]
			for settings_set_name, settings_sets in self._settings.items()
		}

	@property
	def settings_as_strings(self):
		'''
		Return the complete list of sets of settings to use, as strings.
		Take into account the `only_if` parameter.

		Returns
		-------
		settings : dict
			Settings, generated according to their pattern and organized by sets.
		'''

		return {
			set_name: [
				[str(s) for s in settings_set if s.shouldBeDisplayed()]
				for settings_set in settings_sets
			]
			for set_name, settings_sets in self._settings.items()
		}

	@property
	def globalsettings(self):
		'''
		Return the list of global settings, as a name: value dictionary.

		Returns
		-------
		settings : dict
			The global settings.
		'''

		return {setting.name: setting.value for setting in self._globalsettings}

	@property
	def command_line(self):
		'''
		Return the command line to use to generate this simulation.

		Returns
		-------
		command_line : str
			The command line to execute.
		'''

		return ' '.join([self._folder.settings['exec']] + sum(sum(self.settings_as_strings.values(), []), []))

	@property
	def _setting_tag_regex(self):
		'''
		Regex to detect whether there is a setting or global setting tag in a string.

		Returns
		-------
		regex : re.Pattern
			The setting tag regex.
		'''

		if self._setting_tag_regex_compiled is None:
			self._setting_tag_regex_compiled = re.compile(r'\{(?P<category>(?:global)?setting)(?:\[(?P<setname>.+?)\])?(?:\((?P<index>[0-9]+)\))?:(?P<name>.+?)\}')

		return self._setting_tag_regex_compiled

	@property
	def _eval_tag_regex(self):
		'''
		Regex to detect whether we need to evaluate a part of a string.

		Returns
		-------
		regex : re.Pattern
			The eval tag regex.
		'''

		# The `\)*` part is needed so we don't have troubles with functions calls right before the end of a string.
		# Without this, such calls (e.g. in "((sqrt(16)))") end the string prematurely.

		if self._eval_tag_regex_compiled is None:
			self._eval_tag_regex_compiled = re.compile(r'\(\((.*?\)*)\)\)')

		return self._eval_tag_regex_compiled

	def generateGlobalSettings(self):
		'''
		Generate the full list of global settings.
		'''

		self._raw_globalsettings = []

		for setting in self._folder.settings['globalsettings']:
			try:
				setting_value = self._user_settings[setting['name']]

			except KeyError:
				setting_value = setting['default']

			self._raw_globalsettings.append(SimulationGlobalSetting(self, setting['name'], setting_value))

	def getSettingCount(self, setting_name, set_name = None):
		'''
		Get the total number of uses of a given setting, globally or locally.

		Parameters
		----------
		setting_name : str
			Name of the setting to get the count of.

		set_name : str
			Name of the set for the local count.

		Raises
		------
		SettingsSetNotFoundError
			The provided set name does not exist.

		SettingNotFoundError
			The provided setting name does not exist.

		Returns
		-------
		count : int
			Local count if `set_name` is not `None`, global count otherwise.
		'''

		try:
			settings = self._indexed_settings['global'] if set_name is None else self._indexed_settings['local'][set_name]

		except KeyError:
			raise SettingsSetNotFoundError(set_name)

		else:
			try:
				return len(settings[setting_name])

			except KeyError:
				raise SettingNotFoundError(set_name, setting_name)

	def _indexSetting(self, setting, set_name):
		'''
		Add a setting to the index.

		Parameters
		----------
		setting : SimulationSetting
			The setting to index.

		set_name : str
			The name of the set the setting belongs to.
		'''

		if not(set_name in self._indexed_settings['local']):
			self._indexed_settings['local'][set_name] = {}

		indexes = []

		for indexes_dict in [self._indexed_settings['global'], self._indexed_settings['local'][set_name]]:
			try:
				indexes_dict[setting.name].append(setting)

			except KeyError:
				indexes_dict[setting.name] = [setting]

			indexes.append(len(indexes_dict[setting.name]) - 1)

		setting.setIndexes(*indexes)

	def _addRawSettingsSet(self, set_name, default_settings, values_set = {}):
		'''
		Add a raw settings set to the list.

		Parameters
		----------
		set_name : str
			Name of the set.

		default_settings : list
			List of the default settings of this set.

		values_set : dict
			Values to override the defaults.
		'''

		set_to_add = copy.deepcopy(default_settings)

		for setting in set_to_add:
			self._indexSetting(setting, set_name)

			try:
				setting.value = values_set[setting.name]

			except KeyError:
				pass

		try:
			self._raw_settings[set_name].append(set_to_add)

		except KeyError:
			self._raw_settings[set_name] = [set_to_add]

	def generateSettings(self):
		'''
		Generate the full list of settings, taking into account the user settings and the default values in the folder.
		The "raw settings" are generated.
		Each set of settings is a list of all settings in this set.
		'''

		if type(self._user_settings['settings']) is list:
			user_settings = {
				setting_set_name: [s['settings'] for s in self._user_settings['settings'] if s['set'] == setting_set_name]
				for setting_set_name in set([s['set'] for s in self._user_settings['settings']])
			}

		else:
			user_settings = self._user_settings['settings']

		default_pattern = self._folder.settings['setting_pattern']

		self._raw_settings = {}
		self._indexed_settings = {'global': {}, 'local': {}}

		for settings_set in self._folder.settings['settings']:
			default_settings = [
				SimulationSetting(self, settings_set['set'], s['name'])
				for s in settings_set['settings']
			]

			try:
				values_sets = user_settings[settings_set['set']]

			except KeyError:
				if settings_set['required']:
					self._addRawSettingsSet(settings_set['set'], default_settings)

			else:
				if not(type(values_sets) is list):
					values_sets = [values_sets]

				for values_set in values_sets:
					self._addRawSettingsSet(settings_set['set'], default_settings, values_set)

	def getSettingValueFromTag(self, match):
		'''
		Retrieve the value of a setting from a setting tag.

		Parameters
		----------
		match : re.Match
			Match object corresponding to a setting tag.

		Raises
		------
		SettingTagNotRecognizedError
			The setting tag has not been recognized/does not refer to an existing setting.

		Returns
		-------
		setting : mixed
			The value of the setting.
		'''

		try:
			if match.group('category') == 'globalsetting':
				return [
					setting.value
					for setting in self._globalsettings
					if setting.name == match.group('name')
				][0]

			set_dict = self._indexed_settings['global'] if match.group('setname') is None else self._indexed_settings['local'][match.group('setname')]
			set_list = set_dict[match.group('name')]

			k = 0 if match.group('index') is None else int(match.group('index'))

			return set_list[k].value

		except:
			raise SettingTagNotRecognizedError(match.group(0))

	def replaceSettingTag(self, match):
		'''
		Replace a setting tag (`{setting[set_name](k):setting_name}`) by the value of the right setting.
		To be called by `re.sub()`.

		Parameters
		----------
		match : re.Match
			Match object corresponding to a setting tag.

		Returns
		-------
		setting_value : str
			The value of the setting.
		'''

		try:
			return str(self.getSettingValueFromTag(match))

		except SettingTagNotRecognizedError:
			return match.group(0)

	def parseString(self, s):
		'''
		Parse a string to take into account possible settings.
		The tag `{setting:name}` is replaced by the value of the simulation's setting named `name`.
		The tag `{globalsetting:name}` is replaced by the value of the global setting named `name`.

		Tags are replaced recursively.

		Parameters
		----------
		s : str
			The string to parse.

		Returns
		-------
		parsed : mixed
			The parsed string, or a copy of the setting if the whole string is just one tag.
		'''

		# If the string is not a string, we don't have anything to do (seems reasonable!)
		# Then, we test if the string represents a number, so we can cast it

		if not(type(s) is str):
			return s

		try:
			return float(s)

		except ValueError:
			pass

		s = s.strip()

		# We search for settings tags in the string, and recursively replace them

		if self._raw_settings is None:
			self.generateSettings()

		fullmatch = self._setting_tag_regex.fullmatch(s)

		if fullmatch:
			try:
				return copy.deepcopy(self.getSettingValueFromTag(fullmatch))

			except KeyError:
				return s

		parsed = self._setting_tag_regex.sub(self.replaceSettingTag, s)

		self._parser_recursion_stack.append(s)

		if not(parsed in self._parser_recursion_stack):
			return self.parseString(parsed)

		self._parser_recursion_stack.clear()

		# Final step: we try to evaluate the needed parts of the string to apply allowed operations, if any.
		# ValueError is raised if the string contains any unallowed operation, like the use of exec() or other evil functions.

		try:
			fullmatch = self._eval_tag_regex.fullmatch(parsed)

			if fullmatch:
				parsed = string.safeEval(fullmatch.group(1))

			else:
				parsed = self._eval_tag_regex.sub(lambda m: str(string.safeEval(m.group(1))), parsed)

		except (SyntaxError, ValueError):
			pass

		return parsed

class SimulationBaseSetting(abc.ABC):
	'''
	Represent a setting, either global or normal (abstract class).

	Parameters
	----------
	simulation : Simulation
		The simulation using this setting.

	setting_name : str
		Name of the setting.

	setting_value : mixed
		Default value of the setting.
	'''

	def __init__(self, simulation, setting_name, setting_value):
		self._simulation = simulation

		self._name = setting_name
		self._value = setting_value

	def __deepcopy__(self, memo):
		'''
		Override the default behavior of `deepcopy()` to keep the references to the Simulation.
		'''

		cls = self.__class__
		result = cls.__new__(cls)
		memo[id(self)] = result

		for k, v in self.__dict__.items():
			if k == '_simulation':
				setattr(result, k, v)

			else:
				setattr(result, k, copy.deepcopy(v, memo))

		return result

	@property
	def name(self):
		'''
		Getter for the setting name.

		Returns
		-------
		name : str
			Name of the setting.
		'''

		return self._name

	@property
	def value(self):
		'''
		Getter for the setting value, with fixers applied.

		Returns
		-------
		value : mixed
			Value of the setting.
		'''

		try:
			fixers = self._fixers

		except AttributeError:
			fixers = {}

		return self._simulation.folder.applyFixers(self._simulation.parseString(self._value), **fixers)

	@value.setter
	def value(self, new_value):
		'''
		Setter for the setting value.

		Parameters
		----------
		new_value : mixed
			New value of the setting
		'''

		self._value = new_value

class SimulationGlobalSetting(SimulationBaseSetting):
	'''
	Represent a global setting.
	'''

	pass

class SimulationSetting(SimulationBaseSetting):
	'''
	Represent a simulation setting.

	Parameters
	----------
	simulation : Simulation
		Simulation that uses this setting.

	set_name : str
		Name of the set this setting is part of.

	setting_name : str
		Name of this setting.

	Raises
	------
	SettingsSetNotFoundError
		The settings set has not been found.

	SettingNotFoundError
		The setting has not been found in the set.
	'''

	def __init__(self, simulation, set_name, setting_name):
		try:
			self._settings_set_dict = [
				settings_set
				for settings_set in simulation.folder.settings['settings']
				if settings_set['set'] == set_name
			][0]

		except IndexError:
			raise SettingsSetNotFoundError(set_name)

		try:
			self._setting_dict = [
				setting
				for setting in self._settings_set_dict['settings']
				if setting['name'] == setting_name
			][0]

		except IndexError:
			raise SettingNotFoundError(set_name, setting_name)

		super().__init__(simulation, setting_name, self._setting_dict['default'])

		self._set_name = set_name
		self._exclude_for_db = 'exclude' in self._setting_dict and self._setting_dict['exclude']
		self._pattern = self._setting_dict['pattern'] if 'pattern' in self._setting_dict else self._simulation.folder.settings['setting_pattern']

		self._use_only_if = 'only_if' in self._setting_dict
		if self._use_only_if:
			self._only_if_value = self._setting_dict['only_if']

		self._fixers_dict = None
		self._namers_dict = None

	def __str__(self):
		'''
		Return a string representing the setting and its value, according to the pattern.

		Returns
		-------
		setting : str
			The representation of the setting.
		'''

		value = self.value

		if type(value) is str and (not(value) or re.search(r'\s', value) is not None):
			value = repr(value)

		elif type(value) is list:
			value = ' '.join(map(str, value))

		return self._pattern.format(name = self.display_name, value = value)

	def as_dict(self):
		'''
		Dictionary representation of the setting.

		Returns
		-------
		setting : dict
			A dictionary listing some properties of the setting.
		'''

		return {
			'name': self.name,
			'value': self.value,
			'local_index': self._local_index,
			'local_total': self._simulation.getSettingCount(self.name, self._set_name),
			'global_index': self._global_index,
			'global_total': self._simulation.getSettingCount(self.name)
		}

	def setIndexes(self, global_index, local_index):
		'''
		Define the global and local indexes of this setting.

		Parameters
		----------
		global_index : int
			Global index of this setting.

		local_index : int
			Local index of this setting.
		'''

		self._global_index = global_index
		self._local_index = local_index

	def _setModifier(self, modifier):
		'''
		Set a "modifier" property (fixers or namers).

		Parameters
		----------
		modifier : str
			Modifier to define.

		Returns
		-------
		modifier_functions : dict
			The functions to call, in the right order.
		'''

		modifier_functions = {'before': [], 'after': []}

		keys_to_search = {
			'before': [
				(self._setting_dict, ''),
				(self._settings_set_dict, ''),
				(self._setting_dict, '_before'),
				(self._settings_set_dict, '_before'),
				(self._setting_dict, '_between_before')
			],
			'after': [
				(self._setting_dict, '_between_after'),
				(self._settings_set_dict, '_after'),
				(self._setting_dict, '_after')
			]
		}

		for when, keys_path in keys_to_search.items():
			for dict_to_search, key_to_search in keys_path:
				try:
					modifier_functions[when] += dict_to_search[f'{modifier}{key_to_search}']

				except KeyError:
					pass

		return modifier_functions

	@property
	def _fixers(self):
		'''
		Get the fixers to apply to the value.

		Returns
		-------
		fixers : dict
			The list of fixers, in the right order.
		'''

		if self._fixers_dict is None:
			self._fixers_dict = self._setModifier('fixers')

		return self._fixers_dict

	@property
	def _namers(self):
		'''
		Get the namers to apply to the name.

		Returns
		-------
		namers : dict
			The list of namers, in the right order.
		'''

		if self._namers_dict is None:
			self._namers_dict = self._setModifier('namers')

		return self._namers_dict

	@property
	def display_name(self):
		'''
		Get the name of the setting to use inside the simulation.

		Returns
		-------
		name : str
			Name to use.
		'''

		return self._simulation.folder.applyNamers(self.as_dict(), **self._namers)

	@property
	def exclude(self):
		'''
		Return `True` if the setting should be excluded to define a simulation.

		Returns
		-------
		exclude : bool
			Exclusion parameter.
		'''

		return self._exclude_for_db

	def shouldBeDisplayed(self):
		'''
		Return `True` if the setting can be displayed/used for the execution.
		This check is based on the `only_if` parameter. There are several cases for the test string.

		1. It is a full test (e.g. `a < b`): we execute it as it is.
		2. It begins by a conditional operator (<, <=, >, >=, ==, !=, in): we test against the setting value.
		3. It does not contain any conditional operator: we test the equality against the setting value.

		Returns
		-------
		should_be_displayed : bool
			The result of the check.
		'''

		if not(self._use_only_if):
			return True

		if type(self._only_if_value) is str:
			first_operator_match = re.search(r'([<>]=?|[!=]=|in)', self._only_if_value.strip())

			if first_operator_match:
				if first_operator_match.start() > 0:
					return self._simulation.parseString(f'(({self._only_if_value}))')

				return self._simulation.parseString(f'(({repr(self.value)} {self._only_if_value}))')

			return self.value == self._simulation.parseString(self._only_if_value)

		return self.value == self._only_if_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .folder import Folder
from .mapper import Mapper
from .errors import *

class Explorer():
	'''
	Use the Mapper to search for particular settings values.

	Parameters
	----------
	mapper : Mapper
		Instance of the Mapper to use.
	'''

	def __init__(self, mapper):
		self._mapper = mapper
		self._found = None

	def _findInNode(self, node, depth = 0, prev_settings = [], prev_values = []):
		'''
		Find true tests in a node.

		Parameters
		----------
		node : dict
			Current node.

		depth : int
			Depth of the node.

		prev_settings : list
			Settings of the parent nodes.

		prev_values : list
			Values of the parent nodes' settings for this node.
		'''

		if depth not in self._found and 'test' in self._tree[depth]:
			self._found[depth] = {
				'settings': prev_settings + node['settings'],
				'test': self._tree[depth]['test'],
				'true_tests': [],
				'evaluations': [],
				'values': []
			}

		for k, m in enumerate(node['map']):
			if m.get('test'):
				self._found[depth]['values'].append(prev_values + m['values'])
				self._found[depth]['evaluations'].append(m['evaluation'])

				sub_map = node['map'][:k+1]
				self._found[depth]['true_tests'].append({
					f'[{j}]': {
						'values': prev_values + sub_map[j]['values'],
						'evaluation': sub_map[j]['evaluation']
					}
					for j in self._tree[depth]['test_requested_indices']
				})

			if 'output' in m:
				self._findInNode(m['output'], depth + 1, prev_settings + node['settings'], prev_values + m['values'])

	def find(self, tree = None):
		'''
		Find the settings leading to `True` tests.

		Parameters
		----------
		tree : dict
			Description of the tree to explore. If `None`, use the latest Mapper output.
		'''

		if tree is not None:
			self._mapper.mapTree(tree)

		self._tree = self._mapper.tree_by_depths

		self._mapper.events.trigger('explorer-find-start')

		self._found = {}
		self._findInNode(self._mapper.output)

		self._mapper.events.trigger('explorer-find-end')

		return self._found

	def _searchInterval(self):
		'''
		Determine the new search interval.
		For the first iteration, use the whole interval.
		For the others, see where the solution should be located.
		'''

		try:
			latest = self._current_search['iterations'][-1]

		except IndexError:
			self._current_search_iteration['interval'] = self._current_search['interval']

		else:
			if self._mapper._test(self._search_tree['tree'], [latest['interval']['evaluations'][0], latest['evaluation']]):
				self._current_search_iteration['interval'] = {
					'bounds': (latest['interval']['bounds'][0], latest['iterate']),
					'evaluations': (latest['interval']['evaluations'][0], latest['evaluation'])
				}

			else:
				self._current_search_iteration['interval'] = {
					'bounds': (latest['iterate'], latest['interval']['bounds'][1]),
					'evaluations': (latest['evaluation'], latest['interval']['evaluations'][1])
				}

	def _dichotomy(self):
		'''
		Calculate the new iterate of a search by using a dichotomy method.
		'''

		self._current_search_iteration['iterate'] = 0.5 * sum(self._current_search_iteration['interval']['bounds'])

	def _secant(self):
		'''
		Calculate the new iterate of a search by using a secant method.
		'''

		x0, x1 = self._current_search_iteration['interval']['bounds']
		y0, y1 = self._current_search_iteration['interval']['evaluations']

		target = self._tree[self._search['depth']]['test_target']

		self._current_search_iteration['iterate'] = x0 + (target - y0) * (x1 - x0) / (y1 - y0)

	def _searchStopCriterion(self):
		'''
		Calculate the stop criterion of the current iteration.
		'''

		if 'test_target' in self._tree[self._search['depth']]:
			self._current_search_iteration['stopping_criterion'] = abs(self._tree[self._search['depth']]['test_target'] - self._current_search_iteration['evaluation'])

		else:
			a, b = self._current_search_iteration['interval']['bounds']
			self._current_search_iteration['stopping_criterion'] = abs(b - a)

	def _buildSearchTree(self):
		'''
		Build the tree to map the new iterate.
		'''

		default_simulation = self._mapper.default_simulation

		for setting, value in zip(self._search['previous_settings'], self._current_search['previous_values']):
			default_simulation[setting['set']][setting['set_index']][setting['name']] = value

		new_tree = self._tree[self._search['depth']]
		new_tree['values'] = [self._current_search_iteration['iterate']]

		self._search_tree = {
			'default': default_simulation,
			'tree': new_tree
		}

	def _searchIteration(self):
		'''
		Iteration of a search.
		'''

		self._mapper.events.trigger('explorer-search-iteration-start')

		self._current_search_iteration = {}
		self._searchInterval()

		if 'test_target' in self._tree[self._search['depth']]:
			self._secant()
		else:
			self._dichotomy()

		self._buildSearchTree()
		self._mapper.mapTree(self._search_tree)

		self._current_search_iteration['evaluation'] = self._mapper.output['map'][0]['evaluation']
		self._searchStopCriterion()

		self._current_search['iterations'].append(self._current_search_iteration)

		self._mapper.events.trigger('explorer-search-iteration-end')

		if self._current_search_iteration['stopping_criterion'] >= self._search_tolerance and len(self._current_search['iterations']) < self._search_itermax:
			self._searchIteration()

	def search(self, tree = None, depth = None, *, tolerance = 1E-5, itermax = 20):
		'''
		Search for the best values to verify a test.
		There must be only one setting at the searched depth.
		The test must involve the last two evaluations.
		The initial values indicated in the map are used to define the search interval: the solution is assumed to be inside it.

		Parameters
		----------
		tree : dict
			Description of the tree to explore. If `None`, use the latest Mapper output.

		depth : int
			Depth to optimize. If `None`, use the first depth defining a test.

		tolerance : float
			Search tolerance.

		itermax : int
			Maximum number of iterations allowed.
		'''

		if self._found is None:
			self.find(tree)

		self._search_tolerance = tolerance
		self._search_itermax = itermax

		if depth is None:
			depth = list(self._found.keys())[0]

		found = self._found[depth]

		self._search = {
			'depth': depth,
			'previous_settings': found['settings'][:-1],
			'setting': found['settings'][-1],
			'searches': []
		}

		self._mapper.events.trigger('explorer-searches-start', self._search, len(found['true_tests']))

		for test in found['true_tests']:
			self._current_search = {
				'previous_values': test['[-1]']['values'][:-1],
				'interval': {
					'bounds': (test['[-2]']['values'][-1], test['[-1]']['values'][-1]),
					'evaluations': (test['[-2]']['evaluation'], test['[-1]']['evaluation'])
				},
				'iterations': []
			}

			self._mapper.events.trigger('explorer-search-start', self._current_search)

			self._searchIteration()

			self._mapper.events.trigger('explorer-search-end')
			self._search['searches'].append(self._current_search)

		self._mapper.events.trigger('explorer-searches-end')

		return self._search



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import os
import stat
import shutil
import subprocess
import paramiko

from . import jsonfiles
from .errors import *

class RemoteFolder():
	'''
	Send files to and receive from a remote folder.

	Parameters
	----------
	folder_conf : dict
		Configuration of the remote folder.

	Raises
	------
	FileNotFoundError
		The configuration file does not exist.
	'''

	def __init__(self, folder_conf):
		self._configuration = folder_conf
		self._local = (self._configuration['host'] == 'local')

	def __enter__(self):
		'''
		Context manager to call `open()` and `close()` automatically.
		'''

		self.open()
		return self

	def __exit__(self, type, value, traceback):
		'''
		Ensure `close()` is called when exiting the context manager.
		'''

		self.close()

	def _connectSSH(self, config):
		'''
		Open an SSH connection.

		Parameters
		----------
		config : dict
			Parameters of the connection to open.
		'''

		ssh = paramiko.SSHClient()
		ssh.load_system_host_keys()
		ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

		connect_params = {'username': config['user']}

		try:
			connect_params['port'] = config['port']

		except KeyError:
			pass

		try:
			gate = self._connectSSH(config['gate'])

		except KeyError:
			pass

		else:
			connect_params['sock'] = gate.get_transport().open_channel('direct-tcpip', (config['host'], 22), ('', 0))

		ssh.connect(config['host'], **connect_params)

		return ssh

	def open(self):
		'''
		Open the connection.
		'''

		if self._local:
			self._sftp = LocalSFTP()

		else:
			self._ssh = self._connectSSH(self._configuration)
			self._sftp = SFTP.from_transport(self._ssh.get_transport())

		if 'working_directory' in self._configuration:
			self._sftp.chdir(self._configuration['working_directory'])

	def close(self):
		'''
		Close the connection.
		'''

		try:
			self._ssh.close()

		except AttributeError:
			pass

	def execute(self, cmd):
		'''
		Remotely execute a command from the working directory.

		Parameters
		----------
		cmd : str
			Command to execute.

		Returns
		-------
		output : paramiko.ChannelFile
			Output of the command (file-like object).
		'''

		if 'working_directory' in self._configuration:
			cmd = f'cd {self._configuration["working_directory"]}; {cmd}'

		if self._local:
			p = subprocess.run(cmd, shell = True, stdout = subprocess.PIPE)

			stdout = io.StringIO(p.stdout.decode())
			return stdout

		else:
			stdin, stdout, stderr = self._ssh.exec_command(cmd)
			return stdout

	def getFileContents(self, remote_path):
		'''
		Retrieve the content of a remote file.

		Parameters
		----------
		remote_path : str
			Path of the remote file to read.

		Returns
		-------
		content : str
			Content of the file, as a string.
		'''

		with self._sftp.open(remote_path, 'r') as f:
			return f.read()

	def putFileContents(self, remote_path, content):
		'''
		Write the content of a remote file.

		Parameters
		----------
		remote_path : str
			Path of the remote file to write.

		content : str
			Content to write.
		'''

		with self._sftp.open(remote_path, 'w') as f:
			f.write(content)

	def appendToFile(self, remote_path, content):
		'''
		Append a string to a remote file.

		Parameters
		----------
		remote_path : str
			Path of the remote file to write into.

		content : str
			Content to append.
		'''

		with self._sftp.open(remote_path, 'a') as f:
			f.write(content)

	def send(self, local_path, remote_path = None, *, delete = False, replace = False):
		'''
		Send a file or a folder.

		Parameters
		----------
		local_path : str
			Path of the file/folder to send.

		remote_path : str
			Path of the remote file/folder to create.

		delete : boolean
			`True` to delete the local file/folder, once sent.

		replace : bool
			If `False`, send a file only if the source is more recent. Otherwise always send it.

		Raises
		------
		FileNotFoundError
			The local file/folder does not exist.

		Returns
		-------
		remote_path : str
			Remote path of the sent file/directory.
		'''

		stats = os.stat(local_path)

		if not(remote_path):
			remote_path = os.path.basename(os.path.normpath(local_path))

		self._sftp.put(local_path, remote_path, replace, delete)

		return remote_path

	def receive(self, remote_path, local_path = None, *, delete = False):
		'''
		Receive (download) a file or a folder.

		Parameters
		----------
		remote_path : str
			Path of the remote file/folder to receive.

		local_path : str
			Name of the file/folder to create.

		delete : boolean
			`True` to delete the remote file/folder.

		Raises
		------
		RemotePathNotFoundError
			The remote file/folder does not exist.

		Returns
		-------
		local_path : str
			Local path of the received file/folder.
		'''

		try:
			stats = self._sftp.stat(remote_path)

		except FileNotFoundError:
			raise RemotePathNotFoundError(remote_path)

		if not(local_path):
			local_path = os.path.basename(os.path.normpath(remote_path))

		self._sftp.get(remote_path, local_path, delete)

		return local_path

	def deleteRemote(self, entries):
		'''
		Recursively delete some remote entries.

		Parameters
		----------
		entries : str|list
			Single path, or list of paths to delete.
		'''

		if type(entries) is not list:
			entries = [entries]

		for entry in entries:
			self._sftp.remove(entry)

	def deleteLocal(self, entries):
		'''
		Recursively delete some local entries.

		Parameters
		----------
		entries : str|list
			Single path, or list of paths to delete.
		'''

		if type(entries) is not list:
			entries = [entries]

		for entry in entries:
			(shutil.rmtree if os.path.isdir(entry) else os.unlink)(entry)

class SFTP(paramiko.SFTPClient):
	'''
	Overwrite some methods to add some options.
	'''

	def _copyLocalChmod(self, local_path, remote_path):
		'''
		Change the chmod of a remote file/folder to reflect a local one.

		Parameters
		----------
		local_path : str
			Name of the file/folder to use to copy the chmod.

		remote_path : str
			Remote path to alter.
		'''

		self.chmod(remote_path, os.stat(local_path).st_mode & 0o777)

	def _copyRemoteChmod(self, remote_path, local_path):
		'''
		Change the chmod of a local file/folder to reflect a remote one.

		Parameters
		----------
		remote_path : str
			Name of the file/folder to use to copy the chmod.

		local_path : str
			Local path to alter.
		'''

		os.chmod(local_path, self.stat(remote_path).st_mode & 0o777)

	def _makedirs(self, directory):
		'''
		Recursively create a directory.

		Parameters
		----------
		directory : str
			Path to create.
		'''

		try:
			self.mkdir(directory)

		except FileNotFoundError:
			self._makedirs(os.path.dirname(os.path.normpath(directory)))
			self.mkdir(directory)

	def put(self, local_path, remote_path, replace = False, delete = False):
		'''
		Send a file or a folder.

		Parameters
		----------
		local_path : str
			Path to the file/folder to send.

		remote_path : str
			Path of the remote file/folder to create.

		replace : bool
			If `False`, send a file only if the source is more recent. Otherwise always send it.

		delete : bool
			If `True`, delete the local file/folder once sent.
		'''

		if os.path.isfile(local_path):
			if not(replace):
				try:
					if os.stat(local_path).st_mtime <= self.stat(remote_path).st_mtime:
						return None

				except FileNotFoundError:
					pass

			try:
				super().put(local_path, remote_path)

			except FileNotFoundError:
				self._makedirs(os.path.dirname(remote_path))
				super().put(local_path, remote_path)

			self._copyLocalChmod(local_path, remote_path)

			if delete:
				os.unlink(local_path)

		else:
			for entry in os.listdir(local_path):
				self.put(os.path.join(local_path, entry), os.path.join(remote_path, entry), replace, delete)

			if delete:
				os.rmdir(local_path)

	def get(self, remote_path, local_path, delete = False):
		'''
		Download a file or a folder.

		Parameters
		----------
		remote_path : str
			Path to the file/folder to download.

		local_path : str
			Path of file/folder to create.

		delete : bool
			If `True`, delete the remote file/folder once downloaded.
		'''

		if stat.S_ISDIR(self.stat(remote_path).st_mode):
			for entry in self.listdir(remote_path):
				self.get(os.path.join(remote_path, entry), os.path.join(local_path, entry), delete)

			if delete:
				self.rmdir(remote_path)

		else:
			try:
				super().get(remote_path, local_path)

			except FileNotFoundError:
				os.makedirs(os.path.dirname(local_path))
				super().get(remote_path, local_path)

			self._copyRemoteChmod(remote_path, local_path)

			if delete:
				self.remove(remote_path)

	def remove(self, path):
		'''
		Remove a file or a folder.

		Parameters
		----------
		path : str
			Path of the file/folder to remove.
		'''

		if stat.S_ISDIR(self.stat(path).st_mode):
			for entry in self.listdir(path):
				self.remove(os.path.join(path, entry))

			self.rmdir(path)

		else:
			super().remove(path)

class LocalSFTP():
	'''
	Implement some actions on local files, with the same methods names than paramiko.SFTP.

	Parameters
	----------
	wd : str
		Working directory (base directory to use for the files).
	'''

	def __init__(self, wd = '.'):
		self._wd = wd

	def chdir(self, wd):
		'''
		Change the working directory.

		Parameters
		----------
		wd : str
			Working directory to use.
		'''

		self._wd = wd

	def path(self, path):
		'''
		Prepend a path with the working directory.

		Parameters
		----------
		path : str
			Path to prepend.

		Returns
		-------
		complete_path : str
			The complete path, prepended.
		'''

		return os.path.join(self._wd, path)

	def put(self, local_path, remote_path, replace = False, delete = False):
		'''
		Copy or move a file or folder into the working directory.

		Parameters
		----------
		local_path : str
			Path to the file/folder to copy.

		remote_path : str
			Path of the copied file/folder.

		replace : bool
			If `False`, send a file only if the source is more recent. Otherwise always send it.

		delete : bool
			If `True`, move the local file/folder, if `False`, copy it.
		'''

		send = shutil.move

		if os.path.isfile(local_path):
			if not(replace):
				try:
					if os.stat(local_path).st_mtime <= os.stat(self.path(remote_path)).st_mtime:
						return None

				except FileNotFoundError:
					pass

			if not(delete):
				send = shutil.copy

		else:
			if not(delete):
				send = shutil.copytree

			if replace and os.path.isdir(self.path(remote_path)):
				shutil.rmtree(self.path(remote_path))

		try:
			send(local_path, self.path(remote_path))

		except FileNotFoundError:
			os.makedirs(os.path.dirname(self.path(remote_path)))
			send(local_path, self.path(remote_path))

	def get(self, remote_path, local_path, delete = False):
		'''
		Copy or move a file or folder from the working directory.

		Parameters
		----------
		remote_path : str
			Path to the file/folder to copy.

		local_path : str
			Path of the copied file/folder.

		delete : bool
			If `True`, move the remote file/folder, if `False`, copy it.
		'''

		if os.path.isdir(local_path):
			shutil.rmtree(local_path)

		receive = shutil.move

		if not(delete):
			receive = shutil.copy if os.path.isfile(self.path(remote_path)) else shutil.copytree

		try:
			receive(self.path(remote_path), local_path)

		except FileNotFoundError:
			os.makedirs(os.path.dirname(local_path))
			receive(self.path(remote_path), local_path)

	def remove(self, path):
		'''
		Remove a file or a folder.

		Parameters
		----------
		path : str
			Path to the file/folder to remove.
		'''

		path = self.path(path)

		if os.path.isfile(path):
			os.unlink(path)

		else:
			shutil.rmtree(path)

	def stat(self, path):
		'''
		Get some numbers about a file.

		Parameters
		----------
		path : str
			Path of the file.

		Returns
		-------
		res : os.stat_result
			Result of the stat function.
		'''

		return os.stat(self.path(path))

	def open(self, path, mode):
		'''
		Open a file.

		Parameters
		----------
		path : str
			Path to the file to open.

		mode : str
			Opening mode.

		Returns
		-------
		fid : TextIOWrapper
			Opened file.
		'''

		return open(self.path(path), mode)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
In this file, the default settings namers are defined.
A namer must admit the following signature.
Convention: prefix the name of the function by `namer_`.

Parameters
----------
setting : dict
	Representation of the setting.

Returns
-------
display_name : str
	The name to use when launching the simulation.
'''

def namer_appendLocalIndex(setting, separator = '-', only_if_multiple = False):
	'''
	Append the local index to the name.

	Parameters
	----------
	only_if_multiple : bool
		Append the index only if the setting is used more than once (locally).

	separator : str
		Separator between the name and the index.
	'''

	if only_if_multiple and setting['local_total'] <= 1:
		return setting['name']

	return setting['name'] + separator + str(setting['local_index'])

def namer_prependLocalIndex(setting, separator = '-', only_if_multiple = False):
	'''
	Prepend the local index to the name.

	Parameters
	----------
	only_if_multiple : bool
		Prepend the index only if the setting is used more than once (locally).

	separator : str
		Separator between the name and the index.
	'''

	if only_if_multiple and setting['local_total'] <= 1:
		return setting['name']

	return str(setting['local_index']) + separator + setting['name']

def namer_appendGlobalIndex(setting, separator = '-', only_if_multiple = False):
	'''
	Append the local index to the name.

	Parameters
	----------
	only_if_multiple : bool
		Append the index only if the setting is used more than once (globally).

	separator : str
		Separator between the name and the index.
	'''

	if only_if_multiple and setting['global_total'] <= 1:
		return setting['name']

	return setting['name'] + separator + str(setting['global_index'])

def namer_prependGlobalIndex(setting, separator = '-', only_if_multiple = False):
	'''
	Prepend the local index to the name.

	Parameters
	----------
	only_if_multiple : bool
		Prepend the index only if the setting is used more than once (globally).

	separator : str
		Separator between the name and the index.
	'''

	if only_if_multiple and setting['global_total'] <= 1:
		return setting['name']

	return str(setting['global_index']) + separator + setting['name']

def namer_suffix(setting, suffix = ''):
	'''
	Append a string to the name.

	Parameters
	----------
	suffix : str
		The string to append.
	'''

	return setting['name'] + suffix

def namer_prefix(setting, prefix = ''):
	'''
	Prepend a string to the name.

	Parameters
	----------
	prefix : str
		The string to prepend.
	'''

	return prefix + setting['name']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import functools

from math import floor, log10

from .errors import *

class UI():
	'''
	Represent the user interface. Make easy the display of lines that can be updated, and the creation of progress bars.

	Parameters
	----------
	progress_bars_length : int
		Length of the progress bars, in characters.

	progress_bars_empty_char : str
		Character to use for the "empty" part of the progress bars.

	progress_bars_full_char : str
		Character to use to fill a progress bar.

	progress_bars_percentage_precision : int
		Precision to use for the display of the percentage in the progress bars.
	'''

	def __init__(self, *, progress_bars_length = 40, progress_bars_empty_char = '░', progress_bars_full_char = '█', progress_bars_percentage_precision = 'auto'):
		self._cursor_vertical_pos = 0
		self._max_line = 0

		self._items = []

		self._progress_bars_length = progress_bars_length
		self._progress_bars_empty_char = progress_bars_empty_char
		self._progress_bars_full_char = progress_bars_full_char
		self._progress_bars_percentage_precision = progress_bars_percentage_precision

	@property
	def _last_line(self):
		'''
		Get the position of the last line.

		Returns
		-------
		pos : int
			The position of the last line.
		'''

		items = sorted(self._items, key = lambda item: item.position)
		if items:
			return items[-1].position + items[-1].height

		return 0

	def moveCursorTo(self, pos):
		'''
		Move the cursor to a new vertical position.

		Parameters
		----------
		pos : int
			The position to go to.
		'''

		cursor_offset = pos - self._cursor_vertical_pos

		if cursor_offset != 0:
			cursor_direction = 'A' if cursor_offset < 0 else 'B'
			print(f'\u001b[{abs(cursor_offset)}{cursor_direction}', end = '\r')

			self._cursor_vertical_pos = pos

	def moveToLastLine(self):
		'''
		Move the cursor to the last line.
		Before, check if the last line already exists.
		'''

		last_line = self._last_line

		if last_line > self._max_line:
			self.moveCursorTo(last_line - 1)
			print('')
			self._cursor_vertical_pos += 1
			self._max_line = last_line

		self.moveCursorTo(last_line)

	def _addItem(self, item_type, args, *, position = -1):
		'''
		Add an item to display.

		Parameters
		----------
		item_type : type
			Type of the item to add (child of UIDisplayedItem).

		args : dict
			Args to pass to the constructor of the item to add.

		position : int
			Position of the item. Default to `-1`: the item is added at the end.

		Returns
		-------
		item : UIDisplayedItem
			The newly added item.
		'''

		if position >= 0:
			self.moveDownFrom(position)
			self.moveCursorTo(position)

		else:
			self.moveToLastLine()

		item = item_type(self, **args)
		self._items.append(item)
		item.render()

		return item

	def addTextLine(self, text, *, position = -1):
		'''
		Add a new text line to display.

		Parameters
		----------
		text : str
			The text to display

		position : int
			Position of the text line. Default to `-1`: the text line is added at the end.

		Returns
		-------
		text_line : UITextLine
			Object representing the added text line.
		'''

		return self._addItem(UITextLine, {
			'text': text
		}, position = position)

	def addProgressBar(self, total, *, bar_length = None, empty_char = None, full_char = None, percentage_precision = None, position = -1):
		'''
		Add a new progress bar.

		Parameters
		----------
		total : int
			The final number to reach to display the famous 100%.

		bar_length : int
			Length of the progress bar.

		empty_char : str
			Character to use for the empty part of the bar.

		full_char : str
			Character to use to fill the bar.

		percentage_precision : int
			Precision to use for the display of the percentage.

		position : int
			Position of the progress bar. Default to `-1`: the progress bar is added at the end.

		Returns
		-------
		progress_bar : UIProgressBar
			Object representing the added progress bar.
		'''

		return self._addItem(UIProgressBar, {
			'total': total,
			'bar_length': bar_length or self._progress_bars_length,
			'empty_char': empty_char or self._progress_bars_empty_char,
			'full_char': full_char or self._progress_bars_full_char,
			'percentage_precision': percentage_precision or self._progress_bars_percentage_precision
		}, position = position)

	def moveUp(self, item):
		'''
		Move an item to the line above, assuming the above line is empty.

		Parameters
		----------
		item : UIDisplayedItem
			Item to move.

		Raises
		------
		UINonMovableLine
			The line can't be moved.
		'''

		if item.position <= 0:
			raise UINonMovableLine(item.position)

		item.clear()
		self.moveCursorTo(item.position - 1)
		item.position -= 1
		item.render()

	def moveUpFrom(self, pos):
		'''
		Move up all lines starting at a given position.

		Parameters
		----------
		pos : int
			Position to start from.

		Raises
		------
		UINonMovableLine
			The line can't be moved.
		'''

		if pos <= 0:
			raise UINonMovableLine(pos)

		items_to_move = [item for item in self._items if item.position >= pos]
		items_to_move.sort(key = lambda item: item.position)

		for item in items_to_move:
			self.moveUp(item)

	def moveDown(self, item):
		'''
		Move an item to the line below, assuming it is empty.

		Parameters
		----------
		item : UIDisplayedItem
			Item to move.

		Raises
		------
		UINonMovableLine
			The line can't be moved.
		'''

		item.clear()
		self.moveCursorTo(item.position + 1)
		item.position += 1
		item.render()

	def moveDownFrom(self, pos):
		'''
		Move down all lines starting at a given position.

		Parameters
		----------
		pos : int
			Position to start from.

		Raises
		------
		UINonMovableLine
			The line can't be moved.
		'''

		items_to_move = [item for item in self._items if item.position >= pos]
		items_to_move.sort(key = lambda item: item.position, reverse = True)

		for item in items_to_move:
			self.moveDown(item)

	def removeItem(self, item):
		'''
		Remove a displayed item, and then move up all the lines below.

		Parameters
		----------
		item : UIDisplayedItem
			The item to remove.
		'''

		item.clear()
		self.moveUpFrom(item.position + 1)
		self._items.remove(item)
		self.moveToLastLine()

class UIDisplayedItem(abc.ABC):
	'''
	Represent an item displayed in the UI (abstract class).

	Parameters
	----------
	ui : UI
		The UI object this item belongs to.
	'''

	def __init__(self, ui):
		self.ui = ui
		self.position = self.ui._cursor_vertical_pos

	@abc.abstractproperty
	def height(self):
		'''
		The number of lines used by the item.

		Returns
		-------
		height : int
			The number of lines.
		'''

		pass

	@abc.abstractproperty
	def width(self):
		'''
		The width, in characters, of the item.

		Returns
		-------
		width : int
			The width of the item.
		'''

		pass

	@classmethod
	def renderer(cls, func):
		'''
		Decorator for children's render() method.
		'''

		@functools.wraps(func)
		def wrapper(self, *args, **kwargs):
			self.ui.moveCursorTo(self.position)

			func(self, *args, **kwargs)

			self.ui.moveToLastLine()

		return wrapper

	@abc.abstractmethod
	def render(self):
		'''
		Render the item.
		'''

		pass

	def clear(self):
		'''
		Display enough spaces to clear the object.
		'''

		self.ui.moveCursorTo(self.position)
		print(' ' * self.width, end = '\r')
		self.ui.moveToLastLine()

class UITextLine(UIDisplayedItem):
	'''
	Represent a text line displayed in the UI.

	Parameters
	----------
	ui : UI
		The UI object this text line belongs to.

	text : str
		The text to display.
	'''

	def __init__(self, ui, text):
		super().__init__(ui)

		self._text = text

	@property
	def height(self):
		'''
		The number of lines used by the text line.
		Currently, always one. Multilines are not supported yet.

		Returns
		-------
		height : int
			The number of lines used by the text.
		'''

		return 1

	@property
	def width(self):
		'''
		The width of the text line, i.e. the length of the text.

		Returns
		-------
		width : int
			The length of the text.
		'''

		return len(self.text)

	@property
	def text(self):
		'''
		Getter for the displayed text.

		Returns
		-------
		text : str
			Displayed text.
		'''

		return self._text

	@UIDisplayedItem.renderer
	def render(self):
		'''
		Print the text.
		'''

		print(self._text, end = '\r')

	@text.setter
	def text(self, new_text):
		'''
		Change the displayed text.

		Parameters
		----------
		new_text : str
			New text to display.
		'''

		self.clear()
		self._text = new_text
		self.render()

class UIProgressBar(UIDisplayedItem):
	'''
	Represent a progress bar displayed in the UI.

	Parameters
	----------
	ui : UI
		The UI object this text line belongs to.

	total : int
		The final number to reach.

	bar_length : int
		Length of the progress bar.

	empty_char : str
		Character to use for the empty part of the bar.

	full_char : str
		Character to use to fill the bar.

	percentage_precision : int|str
		Precision to use for the display of the percentage.
		Special value `'auto'` to guess the needed precision from the total.
	'''

	def __init__(self, ui, total, *, bar_length = 40, empty_char = '░', full_char = '█', percentage_precision = 'auto'):
		super().__init__(ui)

		self._total = total
		self._counter = 0

		self._bar_length = bar_length
		self._empty_char = empty_char
		self._full_char = full_char

		if percentage_precision == 'auto':
			self._percentage_precision = abs(floor(log10(100 / self._total))) if self._total > 100 else 0

		else:
			self._percentage_precision = percentage_precision

		self._pattern = ' '.join([
			f'{{counter:>{len(str(self._total))}d}}/{self._total}',
			f'{{bar:{self._empty_char}<{self._bar_length}}}',
			f'{{percentage:>{5 + self._percentage_precision}.{self._percentage_precision}%}}'
		])

	@property
	def height(self):
		'''
		The number of lines used by the progress bar.

		Returns
		-------
		height : int
			The number of lines used by the progress bar.
		'''

		return 1

	@property
	def width(self):
		'''
		The width of the progress bar.

		Returns
		-------
		width : int
			The complete width (counter + bar + percentage lengths).
		'''

		return len(self._pattern.format(counter = 0, bar = '', percentage = 0))

	@property
	def counter(self):
		'''
		The current value of the counter.

		Returns
		-------
		counter : int
			Counter value.
		'''

		return self._counter

	@UIDisplayedItem.renderer
	def render(self):
		'''
		Print the progress bar.
		'''

		percentage = self._counter / self._total
		n_full_chars = round(percentage * self._bar_length)

		print(self._pattern.format(counter = self._counter, bar = self._full_char * n_full_chars, percentage = percentage), end = '\r')

	@counter.setter
	def counter(self, n):
		'''
		Set the value of the counter.
		'''

		self.clear()
		self._counter = n
		self.render()

	def update(self, delta = 1):
		'''
		Set the value of the counter by adding an increment.

		Parameters
		----------
		delta : int
			Increment to add.
		'''

		self.counter += delta



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import importlib.util
import uuid

def loadModuleFromFile(filename):
	'''
	Load a source file to use as a module.

	Parameters
	----------
	filename : str
		Path to the file to load.

	Returns
	-------
	module : module
		Loaded module.
	'''

	module_name = uuid.uuid4().hex
	spec = importlib.util.spec_from_file_location(module_name, filename)
	module = importlib.util.module_from_spec(spec)
	sys.modules[module_name] = module
	spec.loader.exec_module(module)

	return module



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import codecs
import base64
import hashlib
import uuid
import json
import ast

from math import sqrt, cos, sin, tan, pi

def intOrNone(s):
	'''
	Convert a string into either an integer or a `None`.

	Parameters
	----------
	s : str
		The string to convert.

	Returns
	-------
	converted : int|None
		The wanted integer if `s` represents one, `None` otherwise.
	'''

	if re.match(r'^-?[0-9]+$', s):
		return int(s)

	return None

def fromObject(obj):
	'''
	Represent an object as a base64 string.

	Parameters
	----------
	obj : dict|list
		The object to represent.

	Returns
	-------
	obj_str : str
		The corresponding base64 string.
	'''

	obj_str = json.dumps(obj, sort_keys = True)
	obj_str = base64.urlsafe_b64encode(obj_str.encode('utf-8')).decode('ascii')
	return obj_str

def toObject(s):
	'''
	Convert a base64 string into the object it represents.

	Parameters
	----------
	s : str
		The string to convert.

	Returns
	-------
	obj : dict|list
		The represented object.
	'''

	obj_str = base64.urlsafe_b64decode(s).decode('utf-8')
	return json.loads(obj_str)

def hash(s):
	'''
	Hash a string, mostly to serve as identifier.

	Parameters
	----------
	s : str
		The string to hash.

	Returns
	-------
	hash : str
		The hash.
	'''

	h = hashlib.md5(s.encode('utf-8')).hexdigest()
	b = codecs.encode(codecs.decode(h.encode('utf-8'), 'hex'), 'base64').decode()

	return b[:-1].replace('+', '-').replace('/', '_')[:-2]

def uniqueID():
	'''
	Generate a unique ID.

	Returns
	-------
	id : str
		The generated ID, as a 32-character hexadecimal string.
	'''

	return uuid.uuid4().hex

def plural(n, if_single, if_plural, *, add_n = True):
	'''
	Return a string or another, depending on a number.

	Parameters
	----------
	n : int
		The number to test.

	if_single : str
		The string to return if `n` is lower or equal than 1.

	if_plural : str
		The string to return if `n` is greater than 1.

	add_n : bool
		`True` to prepend the string by the number, `False` to only get the string.

	Returns
	-------
	s : str
		The single or plural string, depending on the test.
	'''

	s = if_plural if n > 1 else if_single

	if add_n:
		s = f'{n} {s}'

	return s

def safeEval(expr):
	'''
	Safely evaluate an expression by allowing only certain tokens. The following is allowed.
	* Numbers (of any kind: 1, 2.3, 5e-4, etc.)
	* Normal strings
	* Lists
	* Booleans
	* Arithmetic operators:
		+, -, *, /, //, %, **
	* Conditional operators:
		==, !=, <, <=, >, >=, and, or, not(), in
	* Mathematical functions:
		abs, sqrt, cos, sin, tan

	Parameters
	----------
	expr : str
		Expression to evaluate.

	Returns
	-------
	res : mixed
		The result of the expression.

	Raises
	------
	SyntaxError
		Raised by ast.parse() when the string cannot be interpreted.

	ValueError
		The expression contains an unallowed token.
	'''

	whitelist_sample = 'False - True <= -1**2 < 1 + f(1) > 1.5 * 2 >= 2 / 2 and 1 // 2 == 0 or not(1 % 2 != 0) or "a" in ["a", "b"]'
	nodes_whitelist = set([x.__class__.__name__ for x in ast.walk(ast.parse(whitelist_sample))])
	names_whitelist = set(['abs', 'sqrt', 'cos', 'sin', 'tan', 'pi'])

	nodes = set([x.__class__.__name__ for x in ast.walk(ast.parse(expr))])
	names = set([x.id for x in ast.walk(ast.parse(expr)) if x.__class__.__name__ == 'Name'])

	if (nodes - nodes_whitelist) | (names - names_whitelist):
		raise ValueError(f'Unallowed expression `{expr}`')

	return eval(expr)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
In this file, the default variables generators are defined.
A variable generator must admit two parameters and return the result corresponding to these two objects.
Behind the scenes, we use functools.reduce() to apply a generator to the list of objects.

Convention: prefix the name of the function by `generator_`.
'''

def time2seconds(t):
	'''
	Convert a time into a number of seconds.

	Parameters
	----------
	t : int|str
		Number of seconds or string formatted as a time.

	Returns
	-------
	seconds : int
		Total number of seconds.
	'''

	if type(t) is int:
		return t

	t_days = 0

	t_time_str = ''
	t_splitdays = t.split('-')

	try:
		t_time_str = t_splitdays[1]

	except IndexError:
		t_time_str = t_splitdays[0]

	else:
		t_days = int(t_splitdays[0])

	t_time = list(map(int, t_time_str.split(':')))
	t_time += [0] * (3 - len(t_time))

	return 86400 * t_days + 3600 * t_time[0] + 60 * t_time[1] + t_time[2]

def seconds2time(s):
	'''
	Convert a number of seconds into a time string.

	Parameters
	----------
	s : int
		Number of seconds.

	Returns
	-------
	t : str
		String: `D-HH:MM:SS`.
	'''

	decomposition = [s // q % m for m, q in [(60, 1), (60, 60), (24, 3600), (s, 86400)]]
	decomposition.reverse()

	t = ''
	if decomposition[0] > 0:
		t = f'{decomposition[0]}-'

	t += ':'.join(map(lambda d: str(d).zfill(2), decomposition[1:]))

	return t

def generator_sum(a, b):
	'''
	Generate the sum of settings.
	'''

	return a + b

def generator_sumTime(a, b):
	'''
	Generate the sum of two times. The output is a time formatted as `D-HH:MM:SS`. Input can be:
	- a `[D-]HH:MM[:SS]` string,
	- an integer (will be interpreted as a number of seconds).
	'''

	return seconds2time(time2seconds(a) + time2seconds(b))

def generator_first(a, b):
	'''
	Get the first element.
	'''

	return a



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil
import re

from . import string, jsonfiles
from .folder import Folder
from .simulation import Simulation
from .maker import Maker, MakerUI
from .events import Events

class Mapper():
	'''
	Use the Maker to generate simulations and associate them to numeric values.

	Parameters
	----------
	simulations_folder : Folder|str
		The simulations folder. Either a `Folder` instance or a path to a folder with a configuration file.

	config_name : str
		Name of the config to use with the Maker.

	generate_only : bool
		`True` to not add the simulations to the manager.
	'''

	def __init__(self, simulations_folder, config_name = None, *, generate_only = True):
		self._simulations_folder = simulations_folder if type(simulations_folder) is Folder else Folder(simulations_folder)
		self._config_name = config_name

		self._generate_only = generate_only
		self._maker_instance = None

		self._default_simulation = None
		self._tree = None
		self._output = None

		self._simulations_dir = None

		self._index_regex_compiled = None

		self.events = Events([
			'read-start', 'read-end',
			'map-start', 'map-end',
			'node-start', 'node-progress', 'node-end',
			'generate-start', 'generate-end',
			'evaluation-start', 'evaluation-end',
			'explorer-find-start', 'explorer-find-end',
			'explorer-searches-start', 'explorer-searches-end',
			'explorer-search-start', 'explorer-search-end',
			'explorer-search-iteration-start', 'explorer-search-iteration-end'
		])

	def __enter__(self):
		'''
		Context manager to call `close()` at the end.
		'''

		return self

	def __exit__(self, type, value, traceback):
		'''
		Ensure `close()` is called when exiting the context manager.
		'''

		self.close()

	@property
	def maker(self):
		'''
		Return the Maker instance.

		Returns
		-------
		maker : Maker
			Instance currently in used.
		'''

		if self._maker_instance is None:
			self._maker_instance = Maker(self._simulations_folder, self._config_name, override_options = {'generate_only': self._generate_only})

		return self._maker_instance

	def close(self):
		'''
		Properly exit the Maker instance.
		'''

		try:
			self._maker_instance.close()

		except AttributeError:
			pass

		else:
			self._maker_instance = None

	@property
	def tree(self):
		'''
		Get the current tree.

		Returns
		-------
		tree : dict
			The current tree to read.
		'''

		return self._tree

	@property
	def default_simulation(self):
		'''
		Get the current default simulation's settings.

		Returns
		-------
		default_simulation : dict
			The settings of the current default simulation.
		'''

		return self._default_simulation.raw_values_settings

	@property
	def output(self):
		'''
		Get the current output.

		Returns
		-------
		output : dict
			The output of the current tree.
		'''

		return self._output

	@property
	def tree_by_depths(self):
		'''
		Get the current tree, sorted by depths.

		Returns
		-------
		tree : dict
			The current tree to read.
		'''

		if self._tree is None:
			return None

		depth = 0
		tree = {depth: self._tree}

		current_depth = self._tree
		while 'foreach' in current_depth:
			depth += 1
			current_depth = current_depth['foreach']
			tree[depth] = current_depth

		return tree

	@property
	def _index_regex(self):
		'''
		Regex to detect an index in a test.

		Returns
		-------
		regex : re.Pattern
			The index regex.
		'''

		if self._index_regex_compiled is None:
			self._index_regex_compiled = re.compile(r'\[(-?[0-9]+)\]')

		return self._index_regex_compiled

	def _buildValues(self, vdesc):
		'''
		Build the list of values a node will use.
		For values description, we don't calculate the interpolated values now. The Simulation parser will, so we can use settings tags.

		Parameters
		----------
		vdesc : list|dict
			Explicit list, or description. A description is a dictionary with the following keys:
				* `from`: the first value(s),
				* `to`: the last value(s),
				* `n`: the number of values.

		Returns
		-------
		values : list
			The complete list of values for the node.
		'''

		if type(vdesc) is list:
			return [
				v if type(v) is list else [v]
				for v in vdesc
			]

		if type(vdesc['from']) is list:
			return [
				[
					f'(({a} + {k} * ({b} - ({a})) / {vdesc["n"] - 1}))'
					for a, b in zip(vdesc['from'], vdesc['to'])
				]
				for k in range(0, vdesc['n'])
			]

		return [
			[f'(({vdesc["from"]} + {k} * ({vdesc["to"]} - ({vdesc["from"]})) / {vdesc["n"] - 1}))']
			for k in range(0, vdesc['n'])
		]

	def _readTreeNode(self, node):
		'''
		Normalize the way a node is defined, i.e. always use `settings` with a list, and build the values.

		Parameters
		----------
		node : dict
			Description of the node.

		Returns
		-------
		normalized : dict
			The normalized node.
		'''

		settings = node.get('settings') or node.get('setting')
		if type(settings) is not list:
			settings = [settings]

		normalized = {
			'settings': [{'set': '', 'set_index': 0, 'name': '', **coords} for coords in settings],
			'values': self._buildValues(node['values'])
		}

		if 'foreach' in node:
			normalized['foreach'] = self._readTreeNode(node['foreach'])

		try:
			normalized['evaluation'] = node['evaluation']

		except KeyError:
			pass

		try:
			normalized['test'] = node['test']

		except KeyError:
			pass

		else:
			try:
				normalized['test'] = node['test_target']

			except KeyError:
				pass

			if isinstance(normalized['test'], (float, int)):
				normalized['test_target'] = normalized['test']
				normalized['test'] = f'([-2] - {normalized["test"]}) * ([-1] - {normalized["test"]}) <= 0'

			else:
				first_operator_match = re.search(r'([<>]=?|[!=]=|in)', normalized['test'].strip())

				if first_operator_match and first_operator_match.start() == 0:
					normalized['test'] = '[-1] ' + normalized['test']

			normalized['test_requested_indices'] = set(map(int, self._index_regex.findall(normalized['test'])))

		try:
			normalized['stop'] = node['stop']

		except KeyError:
			normalized['stop'] = False

		return normalized

	def _readTree(self, tree):
		'''
		Define the default simulation and normalize the way the tree is defined.

		Parameters
		----------
		tree : dict
			Description of the tree to explore.
		'''

		self.events.trigger('read-start')

		self._default_simulation = Simulation(self._simulations_folder, {'settings': tree.get('default') or {}})
		self._tree = self._readTreeNode(tree['tree'])

		self.events.trigger('read-end')

	def _setSimulations(self, settings, simulations_values):
		'''
		Define the set of simulations to generate.

		Parameters
		----------
		settings : list
			The settings to alter.

		simulations_values : list
			The raw values of these settings.

		Returns
		-------
		simulations : list
			The newly defined simulations.
		'''

		if self._simulations_dir is None:
			self._simulations_dir = self._simulations_folder.tempdir()
			self._simulations = []

		simulations_dir = os.path.join(self._simulations_dir, str(len(os.listdir(self._simulations_dir))))

		simulations = []

		for k, values in enumerate(simulations_values):
			simulation = self._default_simulation.copy()
			simulation['folder'] = os.path.join(simulations_dir, str(k))

			for setting, value in zip(settings, values):
				simulation.getSetting(setting).value = value

			simulations.append(simulation)

		self._simulations += simulations

		return simulations

	def _generateSimulation(self, simulation):
		'''
		Generate a simulation.

		Parameters
		----------
		simulation : Simulation
			The simulation to generate.
		'''

		self.events.trigger('generate-start')

		self.maker.run([simulation])

		self.events.trigger('generate-end')

	def _deleteSimulations(self):
		'''
		Delete the simulations directory.
		'''

		if self._simulations_dir is not None:
			shutil.rmtree(self._simulations_dir)

			self._simulations_dir = None
			self._simulations = None

	def _evaluate(self, node, arg_to_pass):
		'''
		Evaluate the simulations of a node, and store the output.

		Parameters
		----------
		node : dict
			The current node.

		arg_to_pass : Simulation|list
			Either the simulation generated by the node, or the list of simulations.

		Returns
		-------
		evaluation : mixed
			Result of the evaluation, or `None` if there is no evaluation found in the node.
		'''

		if 'evaluation' not in node:
			return None

		self.events.trigger('evaluation-start')

		evaluation = self._simulations_folder.evaluations.call(node['evaluation'], arg_to_pass)

		self.events.trigger('evaluation-end')

		return evaluation

	def _test(self, node, evaluations):
		'''
		Apply a test defined in a node.

		Parameters
		----------
		node : dict
			The current node.

		evaluations : list
			List of evaluations to consider.

		Returns
		-------
		result : mixed
			Result of the test (bool) or `None` if no test has been evaluated.
		'''

		if 'test' not in node:
			return None

		test = node['test']

		# Retrieve the requested indices and first see if the ones which should be different are really different
		# To do that, we "fix" negative indexes so they should be equal to their positive counterparts
		# If they are still negative, it seems that we don't have enough elements in the list: it will be detected in the next part
		# It is important to add the length of the list and to not use the modulo, to be sure we consider different elements

		unique_requested_indices = set(map(lambda k: k if k >= 0 else k + len(evaluations), node['test_requested_indices']))

		if len(unique_requested_indices) < len(node['test_requested_indices']):
			return None

		try:
			test = self._index_regex.sub(lambda m: str(evaluations[int(m.group(1))]), test)

		except IndexError:
			return None

		else:
			return string.safeEval(test)

	def _endNode(self, node, arg_to_evaluate, evaluations, output):
		'''
		Actions to execute at the end of the mapping of a node.
		First, evaluate the generated simulation(s).
		Then, apply the test.

		Parameters
		----------
		node : dict
			The mapped node.

		arg_to_evaluate : Simulation|list
			Either the simulation generated by the node, or the list of simulations.

		evaluations : list
			The current list of evaluations for this node.

		output : dict
			Output of the node.

		Returns
		-------
		stop : bool
			`True` if the latest test is `True` and the node's `stop` is `True`.
		'''

		evaluation = self._evaluate(node, arg_to_evaluate)

		if evaluation is None:
			return False

		evaluations.append(evaluation)
		output['evaluation'] = evaluation

		test = self._test(node, evaluations)

		if test is None:
			return False

		output['test'] = test

		return node['stop'] and test

	def _mapNode(self, node, depth = 0, prev_settings = [], prev_settings_values = [], parent_simulations = []):
		'''
		Prepare the values of the settings in a node, and then either map the children, or generate the simulations before evaluate.

		Parameters
		----------
		node : dict
			The node to map.

		depth : int
			Depth of the node.

		prev_settings : list
			The previous settings from the parent nodes.

		prev_settings_values : list
			The values of the previous settings.

		parent_simulations : list
			The set of simulations to fill for the parent node (will be evaluated).

		Returns
		-------
		output : dict
			Output of the node mapping.
		'''

		self.events.trigger('node-start', depth, node)

		map = []
		output = {'settings': node['settings'], 'map': map}

		evaluations = []

		if 'foreach' in node:
			for values in node['values']:
				simulations = []
				sub_output = self._mapNode(node['foreach'], depth+1, prev_settings + node['settings'], prev_settings_values + values, simulations)
				parent_simulations += simulations

				o = {
					'values': [simulations[0].getSetting(setting).value for setting in node['settings']],
					'output': sub_output
				}

				stop = self._endNode(node, simulations, evaluations, o)
				map.append(o)

				self.events.trigger('node-progress', depth)

				if stop:
					break

		else:
			simulations = self._setSimulations(prev_settings + node['settings'], [prev_settings_values + v for v in node['values']])
			generated_simulations = []

			for simulation in simulations:
				self._generateSimulation(simulation)
				generated_simulations.append(simulation)

				o = {
					'values': [simulation.getSetting(setting).value for setting in node['settings']]
				}

				stop = self._endNode(node, simulation, evaluations, o)
				map.append(o)

				self.events.trigger('node-progress', depth)

				if stop:
					break

			parent_simulations += generated_simulations

		self.events.trigger('node-end', depth)

		return output

	def mapTree(self, tree):
		'''
		Map a tree.

		Parameters
		----------
		tree : dict
			Description of the tree to explore.

		Returns
		-------
		output : dict
			Output of the tree mapping.
		'''

		self._readTree(tree)

		self.events.trigger('map-start')

		self._output = self._mapNode(self._tree)

		self._deleteSimulations()

		self.events.trigger('map-end')

		return self._output

class MapperUI(MakerUI):
	'''
	UI to show the different steps of the Mapper.

	Parameters
	----------
	mapper : Mapper
		Instance of the Mapper from which the events are triggered.
	'''

	def __init__(self, mapper):
		super().__init__(mapper.maker)

		self._mapper = mapper

		self._mapper_state_line = None

		self._nodes_lines = {}
		self._nodes_bars = {}

		self._explorer_searches = None
		self._explorer_searches_n_tests = None
		self._explorer_search = None

		self._mapper.events.addListener('read-start', self._readStart)
		self._mapper.events.addListener('read-end', self._readEnd)
		self._mapper.events.addListener('map-start', self._mapStart)
		self._mapper.events.addListener('map-end', self._mapEnd)
		self._mapper.events.addListener('node-start', self._nodeStart)
		self._mapper.events.addListener('node-progress', self._nodeProgress)
		self._mapper.events.addListener('node-end', self._nodeEnd)
		self._mapper.events.addListener('generate-start', self._generateStart)
		self._mapper.events.addListener('generate-end', self._generateEnd)
		self._mapper.events.addListener('evaluation-start', self._evaluationStart)
		self._mapper.events.addListener('evaluation-end', self._evaluationEnd)
		self._mapper.events.addListener('explorer-find-end', self._explorerFindStart)
		self._mapper.events.addListener('explorer-find-end', self._explorerFindEnd)
		self._mapper.events.addListener('explorer-search-start', self._explorerSearchStart)
		self._mapper.events.addListener('explorer-search-end', self._explorerSearchEnd)
		self._mapper.events.addListener('explorer-searches-start', self._explorerSearchesStart)
		self._mapper.events.addListener('explorer-searches-end', self._explorerSearchesEnd)
		self._mapper.events.addListener('explorer-search-iteration-start', self._explorerSearchIterationStart)
		self._mapper.events.addListener('explorer-search-iteration-end', self._explorerSearchIterationEnd)

		self._mapper.maker.events.addListener('run-end', self._clearMakerState)

	def _updateMapperState(self, state):
		'''
		Text line to display the current state of the Mapper.

		Parameters
		----------
		state : str
			State to display.
		'''

		if self._mapper_state_line is None:
			self._mapper_state_line = self.addTextLine(state, position = 0)

		else:
			self._mapper_state_line.text = state

	def _clearMakerState(self, *args, **kwargs):
		'''
		We don't need the Maker state anymore: we erase it.
		'''

		self._clearState()

	def clearState(self):
		'''
		Clear the Mapper state.
		'''

		if self._mapper_state_line is not None:
			self.removeItem(self._mapper_state_line)
			self._mapper_state_line = None

	def _readStart(self):
		'''
		The interpretation of a tree has started.
		'''

		if self._explorer_searches is None:
			self._updateMapperState('Reading a tree…')

	def _readEnd(self):
		'''
		The tree is read.
		'''

		if self._explorer_searches is None:
			self._updateMapperState('Tree read')

	def _mapStart(self):
		'''
		The mapping of a tree has been started.
		'''

		if self._explorer_searches is None:
			self._updateMapperState('Mapping a tree…')

	def _mapEnd(self):
		'''
		The mapping of a tree has ended.
		'''

		if self._explorer_searches is None:
			self._updateMapperState('Tree mapped')

	def _nodeStart(self, depth, node):
		'''
		The mapping of a node has been started.

		Parameters
		----------
		depth : int
			Depth of the node.

		node : dict
			Description of the node.
		'''

		self._nodes_lines[depth] = self.addTextLine(f'Depth {depth}…', position = 2*depth + 1)
		self._nodes_bars[depth] = self.addProgressBar(len(node['values']), position = 2*depth + 2)

	def _nodeProgress(self, depth):
		'''
		The mapping of a node just progressed.

		Parameters
		----------
		depth : int
			Depth of the node.
		'''

		if depth in self._nodes_bars:
			self._nodes_bars[depth].counter += 1

	def _nodeEnd(self, depth):
		'''
		The mapping of a node has ended.

		Parameters
		----------
		depth : int
			Depth of the node.
		'''

		if depth in self._nodes_lines:
			self.removeItem(self._nodes_bars[depth])
			del self._nodes_bars[depth]

			self.removeItem(self._nodes_lines[depth])
			del self._nodes_lines[depth]

	def _generateStart(self):
		'''
		A set of simulations will be generated.
		'''

		pass

	def _generateEnd(self):
		'''
		A set of simulations has been generated.
		'''

		pass

	def _evaluationStart(self):
		'''
		The evaluation of a simulation has been started.
		'''

		pass

	def _evaluationEnd(self):
		'''
		The evaluation of a simulation has ended.
		'''

		pass

	def _explorerFindStart(self):
		'''
		An Explorer starts to find the tests evaluated to `True`.
		'''

		self._updateMapperState('Finding the true tests…')

	def _explorerFindEnd(self):
		'''
		The search for the tests has ended.
		'''

		self._updateMapperState('Exploration ended')

	def _explorerSearchesStart(self, searches, n_tests):
		'''
		An Explorer starts to search optimal settings.

		Parameters
		----------
		searches : dict
			Description of all the searches.

		n_tests : int
			Number of planned searches.
		'''

		self._explorer_searches = searches
		self._explorer_searches_n_tests = n_tests

		self._updateMapperState(f'Searching for {string.plural(n_tests, "value", "values")}…')

	def _explorerSearchesEnd(self):
		'''
		The search has ended.
		'''

		self._explorer_searches = None
		self._explorer_searches_n_tests = None

		self._updateMapperState('Searches ended')

	def _explorerSearchStart(self, search):
		'''
		An Explorer starts to search optimal settings for a particular test.

		Parameters
		----------
		search : dict
			Description of the starting search.
		'''

		self._explorer_search = search

		self._updateMapperState(f'Search {len(self._explorer_searches["searches"])+1}/{self._explorer_searches_n_tests}…')

	def _explorerSearchEnd(self):
		'''
		The search for one test has ended.
		'''

		self._explorer_search = None

		self._updateMapperState('Search ended')

	def _explorerSearchIterationStart(self):
		'''
		An iteration in a search has started.
		'''

		state = f'Search {len(self._explorer_searches["searches"])+1}/{self._explorer_searches_n_tests}'
		state += f', iteration {len(self._explorer_search["iterations"])+1}'

		if self._explorer_search['iterations']:
			state += f', current criterion: {self._explorer_search["iterations"][-1]["stopping_criterion"]}'

		self._updateMapperState(state)

	def _explorerSearchIterationEnd(self):
		'''
		An iteration in a search has ended.
		'''

		pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import errno
import copy
import tempfile
import glob

from . import utils, string, jsonfiles
from .errors import *
from .fcollection import FCollection
from . import namers as default_namers
from . import fixers as default_fixers
from . import checkers as default_checkers

MAIN_FOLDER = '.hateno'
CONFIG_FOLDER = 'config'
SKELETONS_FOLDER = 'skeletons'
SIMULATIONS_FOLDER = 'simulations'
TMP_FOLDER = 'tmp'

CONF_FILENAME = 'hateno.conf'
SIMULATIONS_LIST_FILENAME = 'simulations.list'
RUNNING_MANAGER_INDICATOR_FILENAME = 'manager.running'

class Folder():
	'''
	Base class for each system needing access to the configuration files of a simulations folder.
	Initialize with the simulations folder and load the settings.

	Parameters
	----------
	folder : str
		The simulations folder. Must contain a settings file.

	Raises
	------
	FileNotFoundError
		No configuration file found in the configuration folder.
	'''

	def __init__(self, folder):
		self._folder = folder
		self._conf_folder_path = os.path.join(self._folder, MAIN_FOLDER)
		self._settings_file = os.path.join(self._conf_folder_path, CONF_FILENAME)
		self._tmp_dir = os.path.join(self._conf_folder_path, TMP_FOLDER)

		if not(os.path.isfile(self._settings_file)):
			raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self._settings_file)

		if not(os.path.isdir(self._tmp_dir)):
			os.makedirs(self._tmp_dir)

		self._settings = None

		self._config_folders_dict = None
		self._configs = {}
		self._skeletons_folders_dict = None
		self._skeletons = {}

		self._program_files = None

		self._namers = None
		self._fixers = None
		self._checkers = None
		self._evaluations = None

	@property
	def folder(self):
		'''
		Return the folder's path.

		Returns
		-------
		path : str
			The path.
		'''

		return self._folder

	@property
	def tmp_folder(self):
		'''
		Return the temporary folder's path.

		Returns
		-------
		path : str
			The path.
		'''

		return self._tmp_dir

	def tempdir(self):
		'''
		Create a temporary directory.

		Returns
		-------
		path : str
			The path to the created folder.
		'''

		return tempfile.mkdtemp(dir = self._tmp_dir)

	@property
	def _config_folders(self):
		'''
		The list of the available configuration folders.

		Returns
		-------
		folders : dict
			A dictionary associating the folders' names to their paths.
		'''

		if self._config_folders_dict is None:
			self._config_folders_dict = {}

			# First, the imported folders, then the local one.
			# In this way, the local configs will always overwrite the imported ones.

			if 'import_config' in self.settings:
				if type(self.settings['import_config']) is not list:
					self.settings['import_config'] = [self.settings['import_config']]

				for import_desc in self.settings['import_config']:
					foldername = import_desc.get('name') or import_desc['config']
					self._config_folders_dict[foldername] = os.path.normpath(os.path.join(self._folder, import_desc['folder'], MAIN_FOLDER, CONFIG_FOLDER, foldername))

			base_folder = os.path.join(self._conf_folder_path, CONFIG_FOLDER)

			if os.path.isdir(base_folder):
				for foldername in os.listdir(base_folder):
					path = os.path.join(base_folder, foldername)

					if os.path.isdir(path):
						self._config_folders_dict[foldername] = path

		return self._config_folders_dict

	def config(self, configname, foldername = None):
		'''
		Get a configuration object.

		Parameters
		----------
		configname : str
			Name of the wanted configuration.

		foldername : str
			Name of the configuration folder. If `None`, use the default config indicated in the configuration file.

		Raises
		------
		NoConfigError
			No configuration folder name given.

		Returns
		-------
		config : dict
			Dictionary stored in the right configuration file.
		'''

		foldername = foldername or self.settings.get('default_config')

		if foldername is None:
			if len(self._config_folders) != 1:
				raise NoConfigError()

			else:
				foldername = list(self._config_folders.keys())[0]

		if foldername not in self._config_folders:
			raise ConfigNotFoundError(foldername)

		if foldername not in self._configs:
			self._configs[foldername] = {}

		if configname not in self._configs[foldername]:
			try:
				self._configs[foldername][configname] = jsonfiles.read(os.path.join(self._config_folders[foldername], f'{configname}.json'))

			except FileNotFoundError:
				self._configs[foldername][configname] = None

		return self._configs[foldername][configname]

	@property
	def _skeletons_folders(self):
		'''
		The list of the available skeletons folders.

		Returns
		-------
		folders : dict
			A dictionary associating the folders' names to their paths.
		'''

		if self._skeletons_folders_dict is None:
			self._skeletons_folders_dict = {}

			if 'import_skeleton' in self.settings:
				if type(self.settings['import_skeleton']) is not list:
					self.settings['import_skeleton'] = [self.settings['import_skeleton']]

				for import_desc in self.settings['import_skeleton']:
					foldername = import_desc.get('name') or import_desc['skeleton']
					self._skeletons_folders_dict[foldername] = os.path.normpath(os.path.join(self._folder, import_desc['folder'], MAIN_FOLDER, SKELETONS_FOLDER, foldername))

			base_folder = os.path.join(self._conf_folder_path, SKELETONS_FOLDER)

			if os.path.isdir(base_folder):
				for foldername in os.listdir(base_folder):
					path = os.path.join(base_folder, foldername)

					if os.path.isdir(path):
						self._skeletons_folders_dict[foldername] = path

		return self._skeletons_folders_dict

	def skeletons(self, foldername):
		'''
		Get the paths to the skeletons files in a given folder.

		Parameters
		----------
		foldername : str
			Name of the skeletons folder.

		Returns
		-------
		paths : dict
			The lists of paths: subgroups skeletons, wholegroup skeletons and script to launch "coordinates".
		'''

		if foldername not in self._skeletons:
			folder = self._skeletons_folders[foldername]
			recipe = jsonfiles.read(os.path.join(folder, 'recipe.json'))

			self._skeletons[foldername] = {
				category: [
					os.path.join(folder, skeleton)
					for skeleton in recipe[category]
				]
				for category in ['subgroups', 'wholegroup']
			}

			option_split = os.path.join(folder, recipe['launch']).rsplit(':', maxsplit = 2)
			option_split_num = [string.intOrNone(s) for s in option_split]

			cut = max([k for k, n in enumerate(option_split_num) if n is None]) + 1

			script_name = ':'.join(option_split[:cut])
			coords = option_split_num[cut:]
			coords += [-1] * (2 - len(coords))

			self._skeletons[foldername]['script_to_launch'] = {
				'name': script_name,
				'coords': coords
			}

		return self._skeletons[foldername]

	@property
	def simulations_list_filename(self):
		'''
		Return the path to the file where the list of simulations is stored.

		Returns
		-------
		path : str
			Path to the simulations list file.
		'''

		return os.path.join(self._conf_folder_path, SIMULATIONS_LIST_FILENAME)

	@property
	def simulations_folder(self):
		'''
		Return the path to the folder where the simulations are stored.
		Create the folder if it does not exist.

		Returns
		-------
		path : str
			Path to the simulations folder.
		'''

		path = os.path.join(self._conf_folder_path, SIMULATIONS_FOLDER)
		if not(os.path.isdir(path)):
			os.makedirs(path)

		return path

	@property
	def running_manager_indicator_filename(self):
		'''
		Return the path to the file indicating the Manager is currently running.

		Returns
		-------
		path : str
			Path to the indicator file.
		'''

		return os.path.join(self._conf_folder_path, RUNNING_MANAGER_INDICATOR_FILENAME)

	@property
	def settings(self):
		'''
		Return the content of the settings file as a dictionary.

		Returns
		-------
		settings : dict
			The folder's settings.
		'''

		if self._settings is None:
			self._settings = jsonfiles.read(self._settings_file)

			if 'namers' not in self._settings:
				self._settings['namers'] = []

			if 'fixers' not in self._settings:
				self._settings['fixers'] = []

		return self._settings

	@property
	def program_files(self):
		'''
		Get the list of the files defined in the configuration file.

		Returns
		-------
		files : list
			A list of files. Each item is a tuple. First item is the local path, second item is the remote one.
		'''

		if self._program_files is None:
			self._program_files = []

			try:
				for path_item in self.settings['files']:
					given_path, dest = path_item if type(path_item) is list else (path_item, '')

					for path in glob.glob(os.path.normpath(os.path.join(self._conf_folder_path, given_path))):
						if os.path.isfile(path):
							self._program_files.append((path, os.path.join(dest, os.path.basename(path))))

						else:
							self._program_files += [(os.path.join(root, file), os.path.join(dest, root, file)) for root, folders, files in os.walk(path) for file in files]

			except KeyError:
				pass

		return self._program_files

	@property
	def fixers(self):
		'''
		Get the list of available values fixers.

		Returns
		-------
		fixers : FCollection
			The collection of values fixers.
		'''

		if self._fixers is None:
			self._fixers = FCollection(filter_regex = r'^fixer_(?P<name>[A-Za-z0-9_]+)$')
			self._fixers.loadFromModule(default_fixers)

			custom_fixers_file = os.path.join(self._conf_folder_path, 'fixers.py')
			if os.path.isfile(custom_fixers_file):
				self._fixers.loadFromModule(utils.loadModuleFromFile(custom_fixers_file))

		return self._fixers

	@property
	def namers(self):
		'''
		Get the list of available namers.

		Returns
		-------
		namers : FCollection
			The collection of namers.
		'''

		if self._namers is None:
			self._namers = FCollection(filter_regex = r'^namer_(?P<name>[A-Za-z0-9_]+)$')
			self._namers.loadFromModule(default_namers)

			custom_namers_file = os.path.join(self._conf_folder_path, 'namers.py')
			if os.path.isfile(custom_namers_file):
				self._namers.loadFromModule(utils.loadModuleFromFile(custom_namers_file))

		return self._namers

	@property
	def checkers(self):
		'''
		Get the list of available checkers.

		Returns
		-------
		checkers : FCollection
			The collection of checkers.
		'''

		if self._checkers is None:
			self._checkers = FCollection(
				categories = ['file', 'folder', 'global'],
				filter_regex = r'^(?P<category>file|folder|global)_(?P<name>[A-Za-z0-9_]+)$'
			)

			self._checkers.loadFromModule(default_checkers)

			custom_checkers_file = os.path.join(self._conf_folder_path, 'checkers.py')
			if os.path.isfile(custom_checkers_file):
				self._checkers.loadFromModule(utils.loadModuleFromFile(custom_checkers_file))

		return self._checkers

	@property
	def evaluations(self):
		'''
		Get the list of available evaluation functions.

		Returns
		-------
		evaluations : FCollection
			The collection of evaluation functions.
		'''

		if self._evaluations is None:
			self._evaluations = FCollection(filter_regex = r'^eval_(?P<name>[A-Za-z0-9_]+)$')

			evaluations_file = os.path.join(self._conf_folder_path, 'evaluations.py')
			if os.path.isfile(evaluations_file):
				self._evaluations.loadFromModule(utils.loadModuleFromFile(evaluations_file))

		return self._evaluations

	def applyFixers(self, value, *, before = [], after = []):
		'''
		Fix a value to prevent false duplicates (e.g. this prevents to consider `0.0` and `0` as different values).
		Each item of a list of fixers is either a fixer's name or a list beginning with the fixer's name and followed by the arguments to pass to the fixer.

		Parameters
		----------
		value : mixed
			The value to fix.

		before : list
			List of fixers to apply before the global ones.

		after : list
			List of fixers to apply after the global ones.

		Returns
		-------
		fixed : mixed
			The same value, fixed.

		Raises
		------
		FixerNotFoundError
			The fixer's name has not been found.
		'''

		value = copy.deepcopy(value)

		for fixer in before + self.settings['fixers'] + after:
			value = self.fixers.call(fixer, value)

		return value

	def applyNamers(self, setting, *, before = [], after = []):
		'''
		Transform the name of a setting before being used in a simulation.

		Parameters
		----------
		setting : dict
			Representation of the setting.

		before : list
			List of namers to apply before the global ones.

		after : list
			List of namers to apply after the global ones.

		Returns
		-------
		name : str
			The name to use.

		Raises
		------
		NamerNotFoundError
			The namer's name has not been found.
		'''

		for namer in before + self.settings['namers'] + after:
			setting['name'] = self.namers.call(namer, setting)

		return setting['name']

	def checkIntegrity(self, simulation):
		'''
		Check the integrity of a simulation.

		Parameters
		----------
		simulation : Simulation
			The simulation to check.

		Returns
		-------
		success : bool
			`True` if the integrity check is successful, `False` otherwise.
		'''

		tree = {}

		for output_entry in ['files', 'folders']:
			tree[output_entry] = []
			checkers_cat = output_entry[:-1]

			if output_entry in self.settings['output']:
				for output in self.settings['output'][output_entry]:
					parsed_name = str(simulation.parseString(output['name']))
					tree[output_entry].append(parsed_name)

					if 'checks' in output:
						for checker in output['checks']:
							if not(self.checkers.call(checker, simulation, parsed_name, category = checkers_cat)):
								return False

		if 'checks' in self.settings['output']:
			for checker in self.settings['output']['checks']:
				if not(self.checkers.call(checker, simulation, tree, category = 'global')):
					return False

		return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import enum
import time
import os
import shutil
import functools
import json

from .errors import *

class JobsManager():
	'''
	Represent a list of jobs to watch the states of.
	'''

	def __init__(self):
		self._jobs = {}

		self._remote_folder = None
		self._linked_file = None

	def linkToFile(self, filename, *, remote_folder = None):
		'''
		Link the manager to a file, to read and write the jobs states.

		Parameters
		----------
		filename : str
			Name of the file to link.

		remote_folder : RemoteFolder
			Remote folder to use, where the file is located.
			Optional, if not provided, local file is assumed.
		'''

		self._remote_folder = remote_folder
		self._linked_file = filename

	def add(self, *names, ignore_existing = False):
		'''
		Add a job to manage.

		Parameters
		----------
		name : str
			Name of the job.

		ignore_existing : bool
			If `True`, no exception is raised if we try to add an existing job.

		Raises
		------
		JobAlreadyExistingError
			The job already exists in the list.
		'''

		for name in names:
			if name in self._jobs and not(ignore_existing):
				raise JobAlreadyExistingError(name)

			self._jobs[name] = Job()

	def delete(self, *names):
		'''
		Remove a job from the list.

		Parameters
		----------
		name : str
			Name of the job.

		Raises
		------
		JobNotFoundError
			The job does not exist.
		'''

		for name in names:
			try:
				del self._jobs[name]

			except KeyError:
				raise JobNotFoundError(name)

	def clear(self):
		'''
		Clear the jobs list.
		'''

		self._jobs.clear()

	def getJob(self, name):
		'''
		Get a job from its name.

		Parameters
		----------
		name : str
			Name of the job.

		Returns
		-------
		job : Job
			The job named `name`.

		Raises
		------
		JobNotFoundError
			The job does not exist.
		'''

		try:
			return self._jobs[name]

		except KeyError:
			raise JobNotFoundError(name)

	def getJobsWithStates(self, states):
		'''
		Get the list of jobs in given states.

		Parameters
		----------
		states : list
			List of `JobState`s to use to filter the jobs.

		Returns
		-------
		jobs : list
			The list of the jobs in these states, as dictionaries.
		'''

		return [{'name': name, **job.__dict__} for name, job in self._jobs.items() if job.state in states]

	def setJobState(self, name, state):
		'''
		Set the state of a job.

		Parameters
		----------
		name : str
			Name of the job to update.

		state : JobState
			New state of the job.

		Raises
		------
		JobNotFoundError
			The job does not exist.
		'''

		try:
			self._jobs[name].state = state

		except KeyError:
			raise JobNotFoundError(name)

		else:
			if not(self._linked_file is None):
				self._appendToFile(name)

	def setJobTotalSteps(self, name, total_steps):
		'''
		Set the total number of steps of a job.

		Parameters
		----------
		name : str
			Name of the job to update.

		total_steps : int
			Number of steps for this job.

		Raises
		------
		JobNotFoundError
			The job does not exist.
		'''

		try:
			self._jobs[name].total_steps = total_steps

		except KeyError:
			raise JobNotFoundError(name)

		else:
			if not(self._linked_file is None):
				self._appendToFile(name)

	def setJobFinishedSteps(self, name, finished_steps):
		'''
		Set the number of finished steps of a job.

		Parameters
		----------
		name : str
			Name of the job to update.

		finished_steps : int
			Number of finished steps for this job.

		Raises
		------
		JobNotFoundError
			The job does not exist.
		'''

		try:
			self._jobs[name].finished_steps = finished_steps

		except KeyError:
			raise JobNotFoundError(name)

		else:
			if not(self._linked_file is None):
				self._appendToFile(name)

	def getFileContent(self):
		'''
		Get the content of the linked file.

		Returns
		-------
		states : dict
			States stored in the file.
		'''

		try:
			if self._remote_folder is None:
				with open(self._linked_file, 'r') as f:
					file = f.read()

			else:
				file = self._remote_folder.getFileContents(self._linked_file)

		except FileNotFoundError:
			return {}

		else:
			states = [json.loads(line) for line in file.splitlines()]
			return functools.reduce(lambda a, b: {**a, **b}, states) if states else {}

	def updateFromFile(self, *, register_unknown = False):
		'''
		Read the states of the registered jobs from the linked file.

		Parameters
		----------
		register_unknown : bool
			If there are non-registered jobs in the file, register them if `True`. Otherwise, just ignore them.
		'''

		states = self.getFileContent()

		if register_unknown:
			self.add(*(set(states.keys()) - set(self._jobs.keys())))

		for job_name, job in self._jobs.items():
			try:
				job.updateFromDict(states[job_name])

			except KeyError:
				pass

	def _appendToFile(self, job_name):
		'''
		Append the state of a job to the linked file.

		Parameters
		----------
		job_name : str
			Name of the job to append.
		'''

		new_line = json.dumps({job_name: self._jobs[job_name].__dict__}) + '\n'

		if self._remote_folder is None:
			with open(self._linked_file, 'a') as f:
				f.write(new_line)

		else:
			self._remote_folder.appendToFile(self._linked_file, new_line)

class Job():
	'''
	Represent a job, i.e. a program executing one or more simulations.
	'''

	def __init__(self):
		self._state = JobState.WAITING

		self._total_steps = 0
		self._finished_steps = 0

	@property
	def __dict__(self):
		'''
		Rewrite the default dict representation, for easier serialization.

		Returns
		-------
		job_dict : dict
			Representation of the job.
		'''

		return {
			'state': self._state.name,
			'total_steps': self._total_steps,
			'finished_steps': self._finished_steps
		}

	@property
	def state(self):
		'''
		Get the current state of the job.

		Returns
		-------
		state : JobState
			State of the job.
		'''

		return self._state

	@state.setter
	def state(self, new_state):
		'''
		Update the state of the job.

		Parameters
		----------
		new_state : JobState
			The new state to store.
		'''

		self._state = new_state

	@property
	def total_steps(self):
		'''
		Get the total number of steps for this job.

		Returns
		-------
		steps : int
			Number of steps.
		'''

		return self._total_steps

	@total_steps.setter
	def total_steps(self, total_steps):
		'''
		Set the total number of steps for this job.

		Parameters
		----------
		total_steps : int
			The total number of steps.
		'''

		self._total_steps = total_steps

	@property
	def finished_steps(self):
		'''
		Get the current number of finished steps.

		Returns
		-------
		steps : int
			Number of finished steps.
		'''

		return self._finished_steps

	@finished_steps.setter
	def finished_steps(self, steps):
		'''
		Update the number of finished steps.

		Parameters
		----------
		steps : int
			The new number of finished steps.
		'''

		self._finished_steps = steps

	@property
	def progress(self):
		'''
		Get the current progress of this job, based on the number of finished and total steps.

		Returns
		-------
		progress : float
			The current progress (finished_steps / total_steps). Returns 0 if no total number of steps is provided.
		'''

		if self._total_steps == 0:
			return 0

		return self._finished_steps / self._total_steps

	def updateFromDict(self, job_dict):
		'''
		Update the state of the job from a dictionary.

		Parameters
		----------
		job_dict : dict
			Dictionary representing the job.

		Raises
		------
		JobStateNotFoundError
			The description of the state has not been found in the dictionary.

		UnknownJobStateError
			The given job state is unknown.
		'''

		try:
			state = job_dict['state']

		except KeyError:
			raise JobStateNotFoundError()

		else:
			try:
				self.state = JobState[state.upper()]

			except KeyError:
				raise UnknownJobStateError(state)

		try:
			self.total_steps = job_dict['total_steps']

		except KeyError:
			pass

		try:
			self.finished_steps = job_dict['finished_steps']

		except KeyError:
			pass

class JobState(enum.Enum):
	'''
	State of a job (waiting, running, ended, etc.).
	'''

	WAITING = enum.auto()
	RUNNING = enum.auto()
	SUCCEED = enum.auto()
	FAILED = enum.auto()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re

'''
In this file, the default "values fixers" are defined.
A fixer always must admit at least one parameter: the value to fix. It must return the fixed value.
Additional parameters are allowed, set in the configurations file.

Convention: prefix the name of the function by `fixer_`.
'''

def fixer_intFloats(value):
	'''
	Converts floats like `2.0` into integers.
	'''

	if type(value) is float and int(value) == value:
		return int(value)

	return value

def fixer_round(value, n_digits):
	'''
	Round all float numbers.

	Parameters
	----------
	n_digits : int
		Number of digits to keep.
	'''

	if type(value) is float:
		return round(value, n_digits)

	return value

def fixer_sortlist(value):
	'''
	Sort a list.
	'''

	if type(value) is list:
		return sorted(value)

	return value

def fixer_list2str(value, separator = ', '):
	'''
	Convert a list into a string.

	Parameters
	----------
	separator : str
		String to place between each list item.
	'''

	if type(value) is list:
		return separator.join(map(str, value))

	return value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import json

from . import utils

def read(filename, *, allow_generator = False):
	'''
	Read a JSON file.

	Parameters
	----------
	filename : str
		Path to the JSON file to read.

	allow_generator : bool
		`True` to allow the use of a Python script as a generator of the object, `False` to allow JSON only.

	Returns
	-------
	obj : dict|list
		The object described in the JSON file.
	'''

	try:
		file_parts = os.path.splitext(filename)
		ext_parts = file_parts[1].split(':', maxsplit = 1)

		args = []
		if len(ext_parts) == 2:
			args = ext_parts[1].split(',')
			filename = ''.join([file_parts[0], ext_parts[0]])

		with open(filename, 'r') as f:
			return json.loads(f.read())

	except json.decoder.JSONDecodeError:
		if not(allow_generator):
			raise

		module = utils.loadModuleFromFile(filename)
		return module.generate(*args)

def write(obj, filename, *, sort_keys = False):
	'''
	Save an object into a JSON file.

	Parameters
	----------
	obj : dict|list
		Object to save.

	filename : str
		Path to the JSON file.

	sort_keys : bool
		`True` to sort the keys before writing the file.
	'''

	dirname = os.path.dirname(filename)
	if dirname and not(os.path.isdir(dirname)):
		os.makedirs(dirname)

	with open(filename, 'w') as f:
		json.dump(obj, f, sort_keys = sort_keys, indent = '\t', separators = (',', ': '))
